# !/usr/bin/env python3
# -*- coding:utf-8 -*-

# @Time    : 2024/5/10 11:28
# @Author  : fanen.lhy
# @Email   : fanen.lhy@antgroup.com
# @FileName: grpc_server_booster.py

from concurrent import futures
import grpc

from agentuniverse.agent_serve.web.rpc.grpc import agentuniverse_service_pb2, \
    agentuniverse_service_pb2_grpc
from agentuniverse.agent_serve.web.rpc.rpc_server import service_run, service_run_async, service_run_result

GRPC_CONFIG = {}

class AgentUniverseService(agentuniverse_service_pb2_grpc.AgentUniverseService):
    """Implementation class of grpc service."""
    def service_run(self, request, context):
        """
        Synchronous invocation of an agent service. Used in rpc implementation.

        Request Args:
            request(`agentuniverse_service_pb2.AgentServiceRequest`):
                service_id(`str`): The id of the agent service.
                params(`str`): A Json String contains agent params passed to service.
                saved(`bool`): Save the request and result into database.
            context: grpc context, needn't pass anything.

        Return:
            Returns agentuniverse_service_pb2.AgentServiceResponse:
            success: This key holds a boolean value indicating the task was
                successfully or not.
            result: This key points to a nested dictionary that includes the
                result of the task.
        """
        service_result = service_run(
            saved=request.saved,
            params=request.params,
            service_id=request.service_id
        )

        return agentuniverse_service_pb2.AgentServiceResponse(
            result=service_result['result'],
            request_id=service_result['request_id'],
            success=service_result['success'],
            message=service_result['message']
        )

    def service_run_async(self, request, context):
        """
        Async invocation of an agent service, return the request id used to
        get result later. Used in rpc implementation.

        Request Args:
            request(`agentuniverse_service_pb2.AgentServiceRequest`):
                service_id(`str`): The id of the agent service.
                params(`str`): A Json String contains agent params passed to service.
                saved(`bool`): Save the request and result into database.
            context: grpc context, needn't pass anything.

        Return:
            Returns agentuniverse_service_pb2.AgentServiceResponse:
            success: This key holds a boolean value indicating the task was
                successfully or not.
            request_id: Stand for a single request taski, can be used in
                service_run_result api to get the result of async task.
        """
        service_result = service_run_async(
            saved=request.saved,
            params=request.params,
            service_id=request.service_id
        )
        return agentuniverse_service_pb2.AgentServiceResponse(
            result=service_result['result'],
            request_id=service_result['request_id'],
            success=service_result['success'],
            message=service_result['message']
        )

    def service_run_result(self, request, context):
        """
        Get the async service result.

        Request Args:
            request(`agentuniverse_service_pb2.AgentResultRequest`):
                request_id(`str`): Request id returned by async run api.
            context: grpc context, needn't pass anything.

        Return:
            Returns agentuniverse_service_pb2.AgentServiceResponse if
            request_id exists in database.
            success: This key holds a boolean value indicating the task was
                successfully or not.
            result: This key points to a nested dictionary that includes the
                result of the task.
        """
        service_result = service_run_result(
            request_id=request.request_id
        )
        return agentuniverse_service_pb2.AgentServiceResponse(
            result=service_result['result'],
            request_id=service_result['request_id'],
            success=service_result['success'],
            message=service_result['message']
        )


def set_grpc_config(configer):
    GRPC_CONFIG["server_port"] = configer.value.get('GRPC', {}).get('server_port', 50051)
    GRPC_CONFIG["max_workers"] = configer.value.get('GRPC', {}).get('max_workers', 10)


def start_grpc_server():
    """Used to start a grpc server, use configer to read grpc server config if
    applied, or use default config of 10 workers and 50051 port."""
    server_port = GRPC_CONFIG.get('server_port', 50051)
    max_workers = GRPC_CONFIG.get('max_workers', 10)
    server = grpc.server(futures.ThreadPoolExecutor(max_workers=max_workers))
    agentuniverse_service_pb2_grpc.add_AgentUniverseServiceServicer_to_server(
        AgentUniverseService(), server
    )
    server.add_insecure_port(f'[::]:{str(server_port)}')
    server.start()
    print(f"AgentUniverse grpc server start at port {str(server_port)}.")
    server.wait_for_termination()



# !/usr/bin/env python3
# -*- coding:utf-8 -*-

# @Time    : 2024/7/25 23:27
# @Author  : wangchongshi
# @Email   : wangchongshi.wcs@antgroup.com
# @FileName: llm_service.py
from typing import List

from agentuniverse.base.component.component_enum import ComponentEnum
from agentuniverse_product.base.constant.llm_constant import LLM_MODEL_NAME
from agentuniverse_product.base.product import Product
from agentuniverse_product.base.product_manager import ProductManager
from agentuniverse_product.service.model.llm_dto import LlmDTO


class LLMService:
    """LLM Service for aU-product."""

    @staticmethod
    def get_llm_list() -> List[LlmDTO]:
        """Get LLM list."""
        res = []
        product_list: List[Product] = ProductManager().get_instance_obj_list()
        if len(product_list) < 1:
            return res
        for product in product_list:
            if product.type == ComponentEnum.LLM.value:
                llm_dto = LlmDTO(nickname=product.nickname, id=product.id)
                llm = product.instance
                llm_dto.temperature = llm.temperature
                llm_dto.model_name = LLM_MODEL_NAME.get(product.id, [])
                res.append(llm_dto)
        return res



from typing import List, Optional, Iterator, Any, AsyncIterator

from langchain_community.chat_models import ChatOllama
from langchain_core.messages import BaseMessage

from agentuniverse.llm.llm import LLM


class OllamaLangchainInstance(ChatOllama):
    llm: LLM = None

    def __init__(self, llm: LLM):
        super().__init__()
        self.llm = llm
        self.model = llm.model_name

    def _create_chat_stream(
            self,
            messages: List[BaseMessage],
            stop: Optional[List[str]] = None,
            **kwargs: Any,
    ) -> Iterator[str]:
        data = self.llm.call(
            messages=self._convert_messages_to_ollama_messages(messages), stop=stop, **kwargs
        )
        for llm_output in data:
            yield llm_output.raw

    async def _acreate_chat_stream(
            self,
            messages: List[BaseMessage],
            stop: Optional[List[str]] = None,
            **kwargs: Any,
    ) -> AsyncIterator[str]:
        data = await self.llm.acall(
                messages=self._convert_messages_to_ollama_messages(messages), stop=stop, **kwargs
        )
        async for llm_output in data:
            yield llm_output.raw



# !/usr/bin/env python3
# -*- coding:utf-8 -*-

# @Time    : 2024/7/22 18:16
# @Author  : fanen.lhy
# @Email   : fanen.lhy@antgroup.com
# @FileName: image_document.py
from typing import Optional, List, Union
from pydantic import Field
from PIL.Image import Image

from .document import Document

class ImageDocument(Document):
    """The basic class for an ImageDocument.

    Attributes:
        origin_image (Optional[Union[Image, str]]): The original image or base64.
        image_embedding (List[float]): Embedding data associated with the image.
        ocr_text (Optional[str]): Text extracted from the image using OCR.
        ocr_text_embedding (Optional[List[float]]): Embedding data associated with the OCR text.
    """

    origin_image: Optional[Union[Image, str]]
    image_embedding: List[float] = Field(default_factory=list)
    ocr_text: Optional[str]
    ocr_text_embedding: Optional[List[float]]


# !/usr/bin/env python3
# -*- coding:utf-8 -*-

# @Time    : 2024/3/14 15:46
# @Author  : jerry.zzw 
# @Email   : jerry.zzw@antgroup.com
# @FileName: component_manager_base.py
import copy
from typing import TypeVar, Generic

from agentuniverse.base.config.application_configer.application_config_manager import ApplicationConfigManager
from agentuniverse.base.component.component_base import ComponentBase
from agentuniverse.base.component.component_enum import ComponentEnum

# 添加类型范型限定
ComponentTypeVar = TypeVar("ComponentTypeVar", bound=ComponentBase)


class ComponentManagerBase(Generic[ComponentTypeVar]):
    """The ComponentManagerBase class, which is used to define the base class of the component manager."""

    def __init__(self, component_type: ComponentEnum):
        """Initialize the ComponentManagerBase."""
        # The component pool map, which is used to store the component instance.
        # _instance_obj_map - Format: {component_instance_name: component_instance_obj}.
        self._instance_obj_map: dict[str, ComponentTypeVar] = {}
        self._component_type: ComponentEnum = component_type

    def register(self, component_instance_name: str, component_instance_obj: ComponentTypeVar):
        """Register the component instance."""
        if component_instance_name in self._instance_obj_map.keys():
            raise ValueError(f"{self._component_type.value} component object instance with name "
                             f"'{component_instance_name}' already exists.")
        self._instance_obj_map[component_instance_name] = component_instance_obj
        if component_instance_obj.default_symbol:
            self._instance_obj_map["__default_instance__"] = component_instance_obj

    def unregister(self, component_instance_name: str):
        """Unregister the component instance abstractmethod."""
        self._instance_obj_map.pop(component_instance_name)

    def get_instance_obj(self, component_instance_name: str,
                         appname: str = None, new_instance: bool = False) -> ComponentTypeVar:
        """Return the component instance object."""
        if component_instance_name == "__default_instance__":
            return self.get_default_instance(new_instance)
        appname = appname or ApplicationConfigManager().app_configer.base_info_appname
        instance_code = f'{appname}.{self._component_type.value.lower()}.{component_instance_name}'
        if new_instance:
            return copy.deepcopy(self._instance_obj_map.get(instance_code))
        return self._instance_obj_map.get(instance_code)

    def get_default_instance(self, new_instance: bool = False) -> ComponentTypeVar:
        """Return the default instance of component."""
        if new_instance:
            return copy.deepcopy(self._instance_obj_map.get("__default_instance__"))
        return self._instance_obj_map.get("__default_instance__")

    def get_instance_name_list(self) -> list[str]:
        """Return the component instance list."""
        return list(self._instance_obj_map.keys())

    def get_instance_obj_list(self) -> list[ComponentTypeVar]:
        """Return the component instance object list."""
        return list(self._instance_obj_map.values())



# !/usr/bin/env python3
# -*- coding:utf-8 -*-

# @Time    : 2024/5/21 17:12
# @Author  : fanen.lhy
# @Email   : fanen.lhy@antgroup.com
# @FileName: sqldb_wrapper.py

import json
from typing import Optional, Sequence, Any, Dict

from langchain_community.utilities.sql_database import SQLDatabase
from sqlalchemy.orm import sessionmaker

from ..base.config.application_configer.application_config_manager import (
    ApplicationConfigManager
)
from ..base.config.component_configer.configers.sqldb_wrapper_config import SQLDBWrapperConfiger
from ..base.component.component_base import ComponentBase
from ..base.component.component_enum import ComponentEnum


class SQLDBWrapper(ComponentBase):
    """A sql DB wrapper based on sqlalchemy and langchain sql database, """

    # Basic attributes of the service class.
    component_type: ComponentEnum = ComponentEnum.SQLDB_WRAPPER
    name: Optional[str] = None
    description: Optional[str] = None
    __sql_database: Optional[SQLDatabase] = None
    db_session: Optional[sessionmaker] = None
    db_wrapper_configer: Optional[SQLDBWrapperConfiger] = None

    class Config:
        arbitrary_types_allowed = True

    def get_instance_code(self) -> str:
        """Generate the full instance code from sql db wrapper name. """
        app_cfg_manager: ApplicationConfigManager = ApplicationConfigManager()
        appname = app_cfg_manager.app_configer.base_info_appname
        return f"{appname}.{self.component_type.value.lower()}.{self.name}"

    def initialize_by_component_configer(self,
                                         db_wrapper_configer: SQLDBWrapperConfiger
                                         ) -> 'SQLDBWrapper':
        """Initialize the SQLDBWrapper by the ComponentConfiger object.

        Args:
            db_wrapper_configer(SQLDBWrapperConfiger): A configer contains service
            basic info.
        Returns:
            SQLDBWrapper: A SQLDBWrapper instance.
        """
        self.name = db_wrapper_configer.name
        self.description = db_wrapper_configer.description
        self.db_wrapper_configer = db_wrapper_configer
        return self

    def run(self, command: str) -> Sequence[Dict[str, Any]]:
        """
        Execute given sql command and return a result sequence.
        """
        return self.sql_database._execute(command=command)


    def run_with_str_return(self, command: str) -> str:
        """
        Execute given sql command and return a str result, intended to be used
        as a part of llm input. If db wrapper's 'max_string_length' property is
        not negative, result str will be truncated.
        """
        return self.sql_database.run(command, fetch="all")

    @property
    def sql_database(self):
        """
        Lazy init, to ensure that database engine can be init correctly in
        separate processes like gunicorn.
        """
        if not self.__sql_database:
            self.db_wrapper_configer.engine_args.setdefault(
                "json_serializer", lambda x: json.dumps(x, ensure_ascii=False)
            )
            self.db_wrapper_configer.sql_database_args.setdefault(
                "sample_rows_in_table_info", 3
            )
            self.db_wrapper_configer.sql_database_args.setdefault(
                "max_string_length", -1
            )
            self.__sql_database = SQLDatabase.from_uri(
                engine_args=self.db_wrapper_configer.engine_args,
                database_uri=self.db_wrapper_configer.db_uri,
                **self.db_wrapper_configer.sql_database_args
            )
        return self.__sql_database


    def get_session(self):
        """
           Get a sqlalchemy session, used for operating with orm.
        """
        if self.db_session:
            return self.db_session
        # Create database engine
        self.db_session = sessionmaker(bind=self.sql_database._engine)
        return self.db_session



# !/usr/bin/env python3
# -*- coding:utf-8 -*-

# @Time    : 2024/10/17 21:01
# @Author  : wangchongshi
# @Email   : wangchongshi.wcs@antgroup.com
# @FileName: planning_agent.py
from agentuniverse.agent.template.planning_agent_template import PlanningAgentTemplate


class PlanningAgent(PlanningAgentTemplate):
    """Planning Agent class."""



# !/usr/bin/env python3
# -*- coding:utf-8 -*-

# @Time    : 2024/8/13 11:31
# @Author  : fanen.lhy
# @Email   : fanen.lhy@antgroup.com
# @FileName: base_router.py

from typing import List, Tuple

from agentuniverse.agent.action.knowledge.rag_router.rag_router import \
    RagRouter
from agentuniverse.agent.action.knowledge.store.query import Query


class BaseRouter(RagRouter):
    def _rag_route(self, query: Query, store_list: List[str]) \
            -> List[Tuple[Query, str]]:
        return [(query, store) for store in store_list]



# !/usr/bin/env python3
# -*- coding:utf-8 -*-

# @Time    : 2024/3/11 16:14
# @Author  : fanen.lhy
# @Email   : fanen.lhy@antgroup.com
# @FileName: general_logger.py

from abc import ABC, abstractmethod
import json
from typing import Literal

import loguru

from agentuniverse.base.context.framework_context_manager import FrameworkContextManager

LOG_LEVEL = Literal[
    "TRACE",
    "DEBUG",
    "INFO",
    "SUCCESS",
    "WARNING",
    "ERROR",
    "CRITICAL"
]


def _get_context_prefix() -> str:
    """Get a dict contains log prefix info from current context and format it
    to a log prefix string."""
    log_context = FrameworkContextManager().get_context("LOG_CONTEXT")
    if log_context:
        json_str = json.dumps(log_context)
        format_context_prefix = '[' + json_str[1:-1] + ']'
        return format_context_prefix
    else:
        return "[default]"


def _get_source_filter(source: str) -> callable:
    """Create a function to filter out specific messages that need to be
    recorded in target loguru sink.

    Args:
        source (`str`):
            Name of the log source, only message from the same source will be
            recorded.

    Returns:
        A callable filter func returns bool value indicating whether the
        message should be logged.
    """

    def source_filter(record) -> bool:
        return record["extra"].get("source") == source

    return source_filter


class Logger(ABC):
    """The basic class of all logger, define all level log functions."""

    @property
    def _logger(self):
        """Logger field"""
        return loguru.logger

    @abstractmethod
    def warn(self, msg, *args, **kwargs):
        """Log warn level message."""
        raise NotImplementedError

    @abstractmethod
    def info(self, msg, *args, **kwargs):
        """Log info level message."""
        raise NotImplementedError

    @abstractmethod
    def error(self, msg, *args, **kwargs):
        """Log error level message."""
        raise NotImplementedError

    @abstractmethod
    def critical(self, msg, *args, **kwargs):
        """Log critical level message."""
        raise NotImplementedError

    @abstractmethod
    def trace(self, msg, *args, **kwargs):
        """Log trace level message."""
        raise NotImplementedError

    @abstractmethod
    def debug(self, msg, *args, **kwargs):
        """Log debug level message."""
        raise NotImplementedError


class GeneralLogger(Logger):
    """General logger class, create a logger with config from config file for
    separate module."""

    def __init__(self,
                 module_name: str,
                 log_path: str,
                 log_format: str,
                 log_rotation: str,
                 log_retention: str,
                 log_level: LOG_LEVEL = "INFO",
                 add_handler: bool = True):
        """Create a new logger instance used by a specific module.

        Args:
            module_name (`str`):
                Name of the module, also used in log file name.
            log_path (`str`):
                Path of the log file.
            log_format (`str`):
                Format of recorded logs.
            log_rotation (`str`):
                Specifies the log rotation policy, controlling when a new log
                file is created. It can be a time period (e.g., "1 week",
                "10 days"), a file size (e.g., "100 MB"), or a function
                returning True when rotation should occur.
            log_retention (`str`):
                Specifies the duration to keep old log files. It can be a time
                span (e.g., "30 days") or a function to filter the files to be
                retained. Files outside this policy are purged.
            log_level (`LOG_LEVEL`, defaults to `"INFO"`):
                Log level, values should be in LOG_LEVEL.
            add_handler (`bool`, defaults to `True`):
                Whether to add a new loguru handler only used to handle log
                recorded by this instance.
        """
        self.module_name = module_name
        self.log_path = log_path
        self.log_format = log_format
        self.log_rotation = log_rotation
        self.log_retention = log_retention
        if add_handler:
            self._add_handler(log_level)

    def update_properties(self, **kwargs):
        """Update logger properties."""
        for key, value in kwargs.items():
            if hasattr(self, key):
                setattr(self, key, value)
            else:
                raise AttributeError(f"{self.__class__.__name__} "
                                     f"has no attribute '{key}'")

    def warn(self, msg, *args, **kwargs):
        self._logger.opt(depth=1).bind(
            source=self.module_name,
            context_prefix=_get_context_prefix()
        ).warning(msg, *args, **kwargs)

    def info(self, msg, *args, **kwargs):
        self._logger.opt(depth=1).bind(
            source=self.module_name,
            context_prefix=_get_context_prefix()
        ).info(msg, *args, **kwargs)

    def error(self, msg, *args, **kwargs):
        self._logger.opt(depth=1).bind(
            source=self.module_name,
            context_prefix=_get_context_prefix()
        ).error(msg, *args, **kwargs)

    def critical(self, msg, *args, **kwargs):
        self._logger.opt(depth=1).bind(
            source=self.module_name,
            context_prefix=_get_context_prefix()
        ).critical(msg, *args, **kwargs)

    def trace(self, msg, *args, **kwargs):
        self._logger.opt(depth=1).bind(
            source=self.module_name,
            context_prefix=_get_context_prefix()
        ).trace(msg, *args, **kwargs)

    def debug(self, msg, *args, **kwargs):
        self._logger.opt(depth=1).bind(
            source=self.module_name,
            context_prefix=_get_context_prefix()
        ).debug(msg, *args, **kwargs)

    def _add_handler(self, log_level: LOG_LEVEL = "INFO"):
        """Add a new loguru log handler, use instance module name to filter out
        message recorded by this instance.

        Args:
            log_level (`str`, defaults to `"INFO"`):
                Log level, values should be in LOG_LEVEL.
        """
        self._logger.add(
            sink=self.log_path,
            level=log_level,
            format=self.log_format,
            rotation=self.log_rotation,
            retention=self.log_retention,
            compression='zip',
            encoding="utf-8",
            enqueue=True,
            filter=_get_source_filter(self.module_name)
        )



# !/usr/bin/env python3
# -*- coding:utf-8 -*-

# @Time    : 2024/8/12 15:44
# @Author  : fanen.lhy
# @Email   : fanen.lhy@antgroup.com
# @FileName: query_keyword_extractor.py

from typing import Optional

from agentuniverse.agent.action.knowledge.doc_processor.doc_processor_manager import \
    DocProcessorManager
from agentuniverse.agent.action.knowledge.query_paraphraser.query_paraphraser import QueryParaphraser
from agentuniverse.agent.action.knowledge.store.document import Document
from agentuniverse.agent.action.knowledge.store.query import Query
from agentuniverse.base.config.component_configer.component_configer import \
    ComponentConfiger


class QueryKeywordExtractor(QueryParaphraser):
    keyword_extractor: Optional[str] = "jieba_keyword_extractor"

    def query_paraphrase(self, origin_query: Query) -> Query:

        keyword_extractor_instance = DocProcessorManager().get_instance_obj(self.keyword_extractor)

        keywords = keyword_extractor_instance.process_docs(
            [Document(text=origin_query.query_str)]
        )[0].keywords

        origin_query.keywords.update(keywords)
        return origin_query

    def _initialize_by_component_configer(self,
                                         query_paraphraser_configer: ComponentConfiger) -> 'QueryParaphraser':
        super()._initialize_by_component_configer(query_paraphraser_configer)
        if hasattr(query_paraphraser_configer, "keyword_extractor"):
            self.keyword_extractor = query_paraphraser_configer.keyword_extractor
        return self



# !/usr/bin/env python3
# -*- coding:utf-8 -*-

# @Time    : 2024/3/22 14:49
# @Author  : wangchongshi
# @Email   : wangchongshi.wcs@antgroup.com
# @FileName: file_reader.py
from pathlib import Path
from typing import Dict, Type, List, Optional

from agentuniverse.agent.action.knowledge.reader.file.docx_reader import DocxReader
from agentuniverse.agent.action.knowledge.reader.file.pdf_reader import PdfReader
from agentuniverse.agent.action.knowledge.reader.file.pptx_reader import PptxReader
from agentuniverse.agent.action.knowledge.reader.reader import Reader
from agentuniverse.agent.action.knowledge.store.document import Document

DEFAULT_FILE_READERS: Dict[str, Type[Reader]] = {
    ".pdf": PdfReader,
    ".docx": DocxReader,
    ".pptx": PptxReader,
}


class FileReader(Reader):
    """The agentUniverse(aU) file reader class.

    FileReader is used to load data from files based on the provided file paths.

    Attributes:
        file_readers (Dict[str, Type[Reader]], optional): The file reader dictionary,
        the key is the suffix of the file, and the value is the specific file reader class.
    """

    file_readers: Dict[str, Type[Reader]] = DEFAULT_FILE_READERS

    def _load_data(self, file_paths: List[Path], ext_info: Optional[Dict] = None) -> List[Document]:
        document_list = []
        for file_path in file_paths:
            file_suffix = file_path.suffix.lower()
            if file_suffix in self.file_readers.keys():
                file_reader = self.file_readers[file_suffix]()
                document_list.extend(file_reader.load_data(file=file_path, ext_info=ext_info))
        return document_list



# !/usr/bin/env python3
# -*- coding:utf-8 -*-

# @Time    : 2024/8/21 11:34
# @Author  : wangchongshi
# @Email   : wangchongshi.wcs@antgroup.com
# @FileName: workflow_config.py
from typing import Optional, List, Any, Union

from pydantic import BaseModel


class NodeOutputParams(BaseModel):
    name: Optional[str] = None
    type: Optional[str] = None
    value: Optional[Any] = None


class InputValueParams(BaseModel):
    type: Optional[str] = None
    content: Optional[Union[List, str]] = None


class NodeInputParams(BaseModel):
    name: Optional[str] = None
    type: Optional[str] = None
    value: Optional[InputValueParams] = None


class NodeInfoParams(BaseModel):
    name: Optional[str] = None
    type: Optional[str] = None
    value: Optional[Any] = None


class ToolNodeInputParams(BaseModel):
    tool_param: Optional[List[NodeInfoParams]] = list()
    input_param: Optional[List[NodeInputParams]] = list()


class KnowledgeNodeInputParams(BaseModel):
    knowledge_param: Optional[List[NodeInfoParams]] = list()
    input_param: Optional[List[NodeInputParams]] = list()


class AgentNodeInputParams(BaseModel):
    agent_param: Optional[List[NodeInfoParams]] = list()
    input_param: Optional[List[NodeInputParams]] = list()


class LLMNodeInputParams(BaseModel):
    llm_param: Optional[List[NodeInfoParams]] = list()
    input_param: Optional[List[NodeInputParams]] = list()


class EndNodeInputParams(BaseModel):
    input_param: Optional[List[NodeInputParams]] = list()
    prompt: Optional[NodeInfoParams] = None


class ConditionParams(BaseModel):
    compare: Optional[str] = None
    left: Optional[NodeInputParams] = None
    right: Optional[NodeInputParams] = None


class ConditionBranchParams(BaseModel):
    name: Optional[str] = None
    conditions: Optional[List[ConditionParams]] = list()


class ConditionNodeInputParams(BaseModel):
    branches: Optional[List[ConditionBranchParams]] = list()



# !/usr/bin/env python3
# -*- coding:utf-8 -*-

# @Time    : 2024/8/30 15:13
# @Author  : fanen.lhy
# @Email   : fanen.lhy@antgroup.com
# @FileName: chroma_hierarchichal_store.py

from typing import List, Any, Optional
from chromadb.api.models.Collection import Collection

from agentuniverse.agent.action.knowledge.embedding.embedding_manager import EmbeddingManager
from agentuniverse.agent.action.knowledge.store.document import Document
from agentuniverse.agent.action.knowledge.store.query import Query
from agentuniverse.agent.action.knowledge.store.chroma_store import ChromaStore
from agentuniverse.base.config.component_configer.component_configer import \
    ComponentConfiger


class ChromaHierarchicalStore(ChromaStore):
    """Object encapsulating the ChromaDB store that has vector search enabled.

    The ChromaStore object provides insert and retrieval capabilities.

    Attributes:
        collection_name (str): The name of the chroma collection to use.
        collection (Collection): A chroma collection object.
        persist_path (Optional[str]): Path to save the chroma database.
    """
    search_depth: int = None
    similarity_top_k_list: Optional[List[int]] = []

    def query(self, query: Query, **kwargs) -> List[Document]:
        """Query the chroma collection with the given query and perform multi-layered search.

        Args:
            query (Query): The query object.
            **kwargs: Arbitrary keyword arguments.

        Note:
            If there is no embedding in the specific query, but the embedding model is configured in the store,
            the embedding data of the query is automatically obtained by the embedding model.

        Returns:
            List[Document]: List of documents retrieved by the query.
        """

        embedding = query.embeddings
        if self.embedding_model is not None and len(embedding) == 0:
            embedding = EmbeddingManager().get_instance_obj(
                self.embedding_model
            ).get_embeddings([query.query_str], text_type="query")[0]

        top_k_ids = []
        current_level_ids = None
        query_result = []

        for depth in range(self.search_depth):
            if len(top_k_ids) > 0:
                all_results = []
                top_k = self.similarity_top_k_list[depth] if len(
                                self.similarity_top_k_list) >= depth + 1 else self.similarity_top_k,
                for parent_id in top_k_ids:
                    # Query for each parent_id
                    if len(embedding) > 0:
                        results = self.collection.query(
                            n_results=top_k,
                            query_embeddings=embedding,
                            where={"hierarchical_parent": parent_id}
                        )
                    else:
                        results = self.collection.query(
                            n_results=top_k,
                            query_texts=[query.query_str],
                            where={"hierarchical_parent": parent_id}
                        )
                    all_results.extend(results)
                sorted_results = sorted(all_results,
                                        key=lambda x: x['distance'])

                # Select the top k results
                query_result = sorted_results[:top_k]

            else:
                filter_condition = {
                   "hierarchical_level": depth
                }
                if len(embedding) > 0:
                    query_result = self.collection.query(
                        n_results=self.similarity_top_k_list[depth] if len(
                            self.similarity_top_k_list) >= depth + 1 else self.similarity_top_k,
                        query_embeddings=embedding,
                        where=filter_condition
                    )
                else:
                    query_result = self.collection.query(
                        n_results=self.similarity_top_k_list[depth] if len(
                            self.similarity_top_k_list) >= depth + 1 else self.similarity_top_k,
                        query_texts=[query.query_str],
                        where=filter_condition
                    )

            documents = self.to_documents(query_result)
            top_k_ids = [doc.id for doc in documents]

        return self.to_documents(query_result)

    def _initialize_by_component_configer(self,
                                          chroma_store_configer: ComponentConfiger) -> 'DocProcessor':
        super()._initialize_by_component_configer(chroma_store_configer)
        if hasattr(chroma_store_configer, "search_depth"):
            self.search_depth = chroma_store_configer.search_depth
        return self



# !/usr/bin/env python3
# -*- coding:utf-8 -*-

# @Time    : 2024/10/10 19:45
# @Author  : wangchongshi
# @Email   : wangchongshi.wcs@antgroup.com
# @FileName: sql_alchemy_memory_storage.py
from abc import abstractmethod
import json
from typing import Optional, List, Any

from pydantic import BaseModel, ConfigDict
from sqlalchemy.orm import declarative_base
from sqlalchemy import Integer, String, DateTime, Text, Column, Index, and_, func

from agentuniverse.agent.memory.memory_storage.memory_storage import MemoryStorage
from agentuniverse.agent.memory.message import Message
from agentuniverse.base.config.component_configer.component_configer import ComponentConfiger
from agentuniverse.database.sqldb_wrapper import SQLDBWrapper
from agentuniverse.database.sqldb_wrapper_manager import SQLDBWrapperManager


class BaseMemoryConverter(BaseModel):
    """ The base class for memory converter used for converting between aU Message and SQLAlchemy model.

    Attributes:
        model_class: The SQLAlchemy model class.
        model_config: The model configuration.
    """

    model_class: Any = None
    model_config = ConfigDict(protected_namespaces=())

    @abstractmethod
    def from_sql_model(self, sql_message: Any) -> Message:
        """Convert a SQLAlchemy model to a Message instance."""
        raise NotImplementedError

    @abstractmethod
    def to_sql_model(self, message: Message) -> Any:
        """Convert a Message instance to a SQLAlchemy model."""
        raise NotImplementedError

    @abstractmethod
    def get_sql_model_class(self) -> Any:
        """Get the SQLAlchemy model class."""
        raise NotImplementedError


def create_memory_model(table_name: str, DynamicBase: Any) -> Any:
    """
    Create a memory model for a given table name.

    Args:
        table_name: The name of the table to use.
        DynamicBase: The base class to use for the model.

    Returns:
        The model class.
    """

    class MemoryModel(DynamicBase):
        """The default memory model for SqlAlchemyMemory."""
        __tablename__ = table_name
        id = Column(Integer, primary_key=True, autoincrement=True)
        session_id = Column(String(100), default='')
        agent_id = Column(String(100), default='')
        source = Column(String(500), default='')
        message = Column(Text)
        gmt_created = Column(DateTime, default=func.now())

        __table_args__ = (
            Index('idx_session_id_source', 'session_id', 'agent_id', 'source'),
            Index('idx_agent_id_source', 'agent_id', 'source'),
            Index('idx_gmt_created', 'gmt_created'),
        )

    return MemoryModel


class DefaultMemoryConverter(BaseMemoryConverter):
    """The default memory converter for SqlAlchemyMemory."""

    def __init__(self, table_name: str, **kwargs: Any):
        super().__init__(**kwargs)
        self.model_class = create_memory_model(table_name, declarative_base())

    def from_sql_model(self, sql_message: Any) -> Message:
        """Convert a SQLAlchemy model to a Message instance."""
        return Message.from_dict({'id': sql_message.id, **json.loads(sql_message.message)})

    def to_sql_model(self, message: Message, session_id: str = None, agent_id: str = None, source: str = None) -> Any:
        """Convert a Message instance to a SQLAlchemy model."""
        return self.model_class(
            session_id=session_id, agent_id=agent_id, source=source,
            message=json.dumps(message.to_dict(), ensure_ascii=False)
        )

    def get_sql_model_class(self) -> Any:
        """Get the SQLAlchemy model class."""
        return self.model_class


class SqlAlchemyMemoryStorage(MemoryStorage):
    """SqlAlchemyMemoryStorage class that stores messages in a SQL database.

    Attributes:
        sqldb_table_name (str): The name of the table to store for the memory.
        sqldb_wrapper_name (str): The name of the SQLDBWrapper to use for the memory.
        memory_converter (BaseMemoryConverter): The memory converter to use for the memory.
        _sqldb_wrapper (SQLDBWrapper): The SQLDBWrapper instance to use for the memory.
    """

    sqldb_table_name: Optional[str] = 'memory'
    sqldb_wrapper_name: Optional[str] = None
    memory_converter: BaseMemoryConverter = None
    _sqldb_wrapper: SQLDBWrapper = None

    def _initialize_by_component_configer(self,
                                          memory_storage_config: ComponentConfiger) -> 'SqlAlchemyMemoryStorage':
        """Initialize the SqlAlchemyMemoryStorage by the ComponentConfiger object.

        Args:
            memory_storage_config(ComponentConfiger): A configer contains sql_alchemy_memory_storage basic info.
        Returns:
            SqlAlchemyMemoryStorage: A SqlAlchemyMemoryStorage instance.
        """
        super()._initialize_by_component_configer(memory_storage_config)
        if getattr(memory_storage_config, 'sqldb_table_name', None):
            self.sqldb_table_name = memory_storage_config.sqldb_table_name
        if getattr(memory_storage_config, 'sqldb_wrapper_name', None):
            self.sqldb_wrapper_name = memory_storage_config.sqldb_wrapper_name
        if self.sqldb_wrapper_name is None:
            raise Exception('`sqldb_wrapper_name` is not set')
        # initialize the memory converter if not set
        if self.memory_converter is None:
            self.memory_converter = DefaultMemoryConverter(self.sqldb_table_name)
        return self

    def _init_db(self) -> None:
        """Initialize the database."""
        self._sqldb_wrapper = SQLDBWrapperManager().get_instance_obj(self.sqldb_wrapper_name)
        if self._sqldb_wrapper is None:
            raise Exception('The sqldb_wrapper for the `sqldb_wrapper_name` was not found,'
                            ' please check the `sqldb_wrapper_name`.')
        self._create_table_if_not_exists()

    def _create_table_if_not_exists(self) -> None:
        """Create the db table if it does not exist."""
        with self._sqldb_wrapper.sql_database._engine.connect() as conn:
            if not conn.dialect.has_table(conn, self.sqldb_table_name):
                self.memory_converter.get_sql_model_class().__table__.create(conn)

    def delete(self, session_id: str = None, agent_id: str = None, **kwargs) -> None:
        """Delete the memory from the database.

        Args:
            session_id (str): The session id of the memory to delete.
            agent_id (str): The agent id of the memory to delete.
        """
        if self._sqldb_wrapper is None:
            self._init_db()
        if session_id is None and agent_id is None:
            return
        with self._sqldb_wrapper.get_session()() as session:
            model_class = self.memory_converter.get_sql_model_class()
            query = session.query(model_class)
            # construct query based on the provided session_id and agent_id
            if session_id is not None:
                query = query.filter(getattr(model_class, 'session_id') == session_id)
            if agent_id is not None:
                query = query.filter(getattr(model_class, 'agent_id') == agent_id)

            # execute delete and commit the session
            query.delete(synchronize_session=False)
            session.commit()

    def add(self, message_list: List[Message], session_id: str = None, agent_id: str = None, **kwargs) -> None:
        """Add messages to the memory db.

        Args:
            message_list (List[Message]): The list of messages to add.
            session_id (str): The session id of the memory to add.
            agent_id (str): The agent id of the memory to add.
        """
        if self._sqldb_wrapper is None:
            self._init_db()
        if message_list is None:
            return
        with self._sqldb_wrapper.get_session()() as session:
            for message in message_list:
                session.add(
                    self.memory_converter.to_sql_model(message=message, session_id=session_id if session_id else None,
                                                       agent_id=agent_id if agent_id else None,
                                                       source=message.source if message.source else None))
            session.commit()

    def get(self, session_id: str = None, agent_id: str = None, top_k=10, source: str = None, **kwargs) -> List[
        Message]:
        """Get messages from the memory db.

        Args:
            session_id (str): The session id of the memory to get.
            agent_id (str): The agent id of the memory to get.
            top_k (int): The number of messages to get.
            source (str): The source of the messages to get.

        Returns:
            List[Message]: The list of messages retrieved from the memory.
        """
        if self._sqldb_wrapper is None:
            self._init_db()
        with self._sqldb_wrapper.get_session()() as session:
            # get the messages from the memory by session_id and agent_id
            model_class = self.memory_converter.get_sql_model_class()
            conditions = []

            # conditionally add session_id to the query
            if session_id:
                session_id_col = getattr(model_class, 'session_id')
                conditions.append(session_id_col == session_id)

            # conditionally add agent_id to the query
            if agent_id:
                agent_id_col = getattr(model_class, 'agent_id')
                conditions.append(agent_id_col == agent_id)

            # conditionally add source to the query
            if source:
                source_col = getattr(model_class, 'source')
                conditions.append(source_col == source)

            # build the query with dynamic conditions
            query = session.query(self.memory_converter.model_class)
            if conditions:
                query = query.where(and_(*conditions))
            query = query.order_by(model_class.gmt_created.asc())

            # Execute the query and fetch the results
            records = query.all()

            records = records[-top_k:]

            messages = []
            for record in records:
                messages.append(self.memory_converter.from_sql_model(record))
            return messages



# !/usr/bin/env python3
# -*- coding:utf-8 -*-

# @Time    : 2024/10/9 19:24
# @Author  : wangchongshi
# @Email   : wangchongshi.wcs@antgroup.com
# @FileName: memory_compressor.py
from typing import Optional, List

from langchain_core.output_parsers import StrOutputParser

from agentuniverse.agent.memory.message import Message
from agentuniverse.base.component.component_base import ComponentEnum
from agentuniverse.base.component.component_base import ComponentBase
from agentuniverse.base.config.component_configer.component_configer import ComponentConfiger
from agentuniverse.base.util.memory_util import get_memory_string, get_memory_tokens
from agentuniverse.llm.llm import LLM
from agentuniverse.llm.llm_manager import LLMManager
from agentuniverse.prompt.prompt import Prompt
from agentuniverse.prompt.prompt_manager import PromptManager


class MemoryCompressor(ComponentBase):
    """The basic class for the memory compressor.

    Attributes:
        name (str): The name of the memory compressor.
        description (str): The description of the memory compressor.
        compressor_prompt_version (str): The version of the prompt used for compressing the memory.
        compressor_llm_name (str): The name of the LLM used for compressing the memory.
    """

    name: Optional[str] = None
    description: Optional[str] = None
    compressor_prompt_version: Optional[str] = None
    compressor_llm_name: Optional[str] = None
    component_type: ComponentEnum = ComponentEnum.MEMORY_COMPRESSOR

    def compress_memory(self, new_memories: List[Message], max_tokens: int = 500, existing_memory: str = '',
                        **kwargs) -> str:
        """Compress the memory.

        Args:
            new_memories (List[Message]): The new memories to compress.
            max_tokens (int): The maximum number of tokens allowed in the compressed memory.
            existing_memory (str): The existing memory to append to.

        Returns:
            str: The compressed memory.
        """
        prompt: Prompt = PromptManager().get_instance_obj(self.compressor_prompt_version)
        llm: LLM = LLMManager().get_instance_obj(self.compressor_llm_name)
        if prompt and llm:
            new_memory_str = get_memory_string(new_memories)
            chain = prompt.as_langchain() | llm.as_langchain() | StrOutputParser()
            return chain.invoke(
                input={'new_lines': new_memory_str, 'summary': existing_memory, 'max_tokens': max_tokens})
        else:
            return ''

    def _initialize_by_component_configer(self, memory_compressor_config: ComponentConfiger) -> 'MemoryCompressor':
        """Initialize the MemoryCompressor by the ComponentConfiger object.

        Args:
            memory_compressor_config(ComponentConfiger): A configer contains memory_compressor basic info.
        Returns:
            MemoryCompressor: A MemoryCompressor instance.
        """
        if getattr(memory_compressor_config, 'name', None):
            self.name = memory_compressor_config.name
        if getattr(memory_compressor_config, 'description', None):
            self.description = memory_compressor_config.description
        if getattr(memory_compressor_config, 'compressor_prompt_version', None):
            self.compressor_prompt_version = memory_compressor_config.compressor_prompt_version
        if getattr(memory_compressor_config, 'compressor_llm_name', None):
            self.compressor_llm_name = memory_compressor_config.compressor_llm_name
        return self



# !/usr/bin/env python3
# -*- coding:utf-8 -*-

# @Time    : 2024/7/24 11:41
# @Author  : fanen.lhy
# @Email   : fanen.lhy@antgroup.com
# @FileName: reader_manager.py

from agentuniverse.base.annotation.singleton import singleton
from agentuniverse.base.component.component_enum import ComponentEnum
from agentuniverse.base.component.component_manager_base import \
    ComponentManagerBase, ComponentTypeVar
from agentuniverse.agent.action.knowledge.reader.reader import Reader


@singleton
class ReaderManager(ComponentManagerBase[Reader]):
    """A singleton manager class of the reader."""
    DEFAULT_READER = {
        "pdf": "default_pdf_reader",
        "pptx": "default_pptx_reader",
        "docx": "default_docx_reader",
        "txt": "default_txt_reader"
    }

    def __init__(self):
        super().__init__(ComponentEnum.READER)

    def get_file_default_reader(self,
                                file_type: str,
                                new_instance: bool = False) -> Reader | None:
        if file_type in self.DEFAULT_READER:
            return self.get_instance_obj(self.DEFAULT_READER[file_type])
        else:
            return None



# !/usr/bin/env python3
# -*- coding:utf-8 -*-

# @Time    : 2024/7/26 14:32
# @Author  : wangchongshi
# @Email   : wangchongshi.wcs@antgroup.com
# @FileName: message_dto.py
from typing import Optional

from pydantic import BaseModel, Field


class MessageDTO(BaseModel):
    id: int = Field(description="ID")
    session_id: str = Field(description="Session ID")
    content: Optional[str] = Field(description="message content", default="")
    gmt_created: Optional[str] = Field(description="message create time")
    gmt_modified: Optional[str] = Field(description="message update time")



# !/usr/bin/env python3
# -*- coding:utf-8 -*-

# @Time    : 2024/9/29 14:41
# @Author  : wangchongshi
# @Email   : wangchongshi.wcs@antgroup.com
# @FileName: reviewing_agent_template.py
from queue import Queue

from langchain_core.utils.json import parse_json_markdown

from agentuniverse.agent.input_object import InputObject
from agentuniverse.agent.template.agent_template import AgentTemplate
from agentuniverse.base.config.component_configer.configers.agent_configer import AgentConfiger
from agentuniverse.base.util.common_util import stream_output
from agentuniverse.base.util.logging.logging_util import LOGGER


class ReviewingAgentTemplate(AgentTemplate):

    def input_keys(self) -> list[str]:
        return ['input', 'expressing_result']

    def output_keys(self) -> list[str]:
        return ['output', 'score', 'suggestion']

    def parse_input(self, input_object: InputObject, agent_input: dict) -> dict:
        agent_input['input'] = input_object.get_data('input')
        agent_input['expressing_result'] = input_object.get_data('expressing_result').get_data('output')
        agent_input['expert_framework'] = input_object.get_data('expert_framework', {}).get('reviewing')
        return agent_input

    def parse_result(self, agent_result: dict) -> dict:
        final_result = dict()

        output = agent_result.get('output')
        output = parse_json_markdown(output)

        is_useful = output.get('is_useful')
        if is_useful is None:
            is_useful = False
        is_useful = bool(is_useful)
        if is_useful:
            score = 80
        else:
            score = 0

        final_result['output'] = output
        final_result['score'] = score
        final_result['suggestion'] = output.get('suggestion')
        # add reviewing agent log info.
        logger_info = f"\nReviewing agent execution result is :\n"
        reviewing_info_str = f"review suggestion: {final_result.get('suggestion')} \n"
        reviewing_info_str += f"review score: {final_result.get('score')} \n"
        LOGGER.info(logger_info + reviewing_info_str)

        return final_result

    def add_output_stream(self, output_stream: Queue, agent_output: str) -> None:
        if not output_stream:
            return
        # add reviewing agent final result into the stream output.
        stream_output(output_stream,
                      {"data": {
                          'output': parse_json_markdown(agent_output).get('suggestion'),
                          "agent_info": self.agent_model.info
                      }, "type": "reviewing"})

    def initialize_by_component_configer(self, component_configer: AgentConfiger) -> 'ReviewingAgentTemplate':
        """Initialize the Agent by the AgentConfiger object.

        Args:
            component_configer(AgentConfiger): the ComponentConfiger object
        Returns:
            ReviewingAgentTemplate: the ReviewingAgentTemplate object
        """
        super().initialize_by_component_configer(component_configer)
        self.prompt_version = self.agent_model.profile.get('prompt_version', 'default_reviewing_agent.cn')
        self.validate_required_params()
        return self

    def validate_required_params(self):
        if not self.llm_name:
            raise ValueError(f'llm_name of the agent {self.agent_model.info.get("name")}'
                             f' is not set, please go to the agent profile configuration'
                             ' and set the `name` attribute in the `llm_model`.')



# !/usr/bin/env python3
# -*- coding:utf-8 -*-

# @Time    : 2024/5/21 13:52
# @Author  : weizjajj
# @Email   : weizhongjie.wzj@antgroup.com
# @FileName: kimi_openai_style_llm.py
from typing import Optional, Any, Union, Iterator, AsyncIterator

import requests
from pydantic import Field

from agentuniverse.base.util.env_util import get_from_env
from agentuniverse.llm.llm_output import LLMOutput
from agentuniverse.llm.openai_style_llm import OpenAIStyleLLM

KIMI_Max_CONTEXT_LENGTH = {
    "moonshot-v1-8k": 8000,
    "moonshot-v1-32k": 32000,
    "moonshot-v1-128k": 128000
}


class KIMIOpenAIStyleLLM(OpenAIStyleLLM):
    """
        KIMI's OpenAI Style LLM
        Attributes:
            api_key (Optional[str]): The API key to use for authentication. Defaults to None.
            api_base (Optional[str]): The base URL to use for the API. Defaults to None.
    """
    api_key: Optional[str] = Field(default_factory=lambda: get_from_env("KIMI_API_KEY"))
    api_base: Optional[str] = "https://api.moonshot.cn/v1"
    proxy: Optional[str] = Field(default_factory=lambda: get_from_env("KIMI_PROXY"))
    organization: Optional[str] = Field(default_factory=lambda: get_from_env("KIMI_ORGANIZATION"))

    def _call(self, messages: list, **kwargs: Any) -> Union[LLMOutput, Iterator[LLMOutput]]:
        """ The call method of the LLM.

        Users can customize how the model interacts by overriding call method of the LLM class.

        Args:
            messages (list): The messages to send to the LLM.
            **kwargs: Arbitrary keyword arguments.
        """
        return super()._call(messages, **kwargs)

    async def _acall(self, messages: list, **kwargs: Any) -> Union[LLMOutput, AsyncIterator[LLMOutput]]:
        """ The async call method of the LLM.

        Users can customize how the model interacts by overriding acall method of the LLM class.

        Args:
            messages (list): The messages to send to the LLM.
            **kwargs: Arbitrary keyword arguments.
        """
        return await super()._acall(messages, **kwargs)

    def max_context_length(self) -> int:
        if super().max_context_length():
            return super().max_context_length()
        return KIMI_Max_CONTEXT_LENGTH.get(self.model_name, 8000)

    def get_num_tokens(self, text: str) -> int:
        # 通过http获取token数量
        messages = [{"role": "user", "content": text}]
        body = {"model": self.model_name, "messages": messages}
        headers = {'Content-Type': 'application/json', 'Authorization': f'Bearer {self.api_key}'}
        res = requests.post(f"{self.api_base}/tokenizers/estimate-token-count", headers=headers, json=body)
        return res.json().get('data').get('total_tokens')



# !/usr/bin/env python3
# -*- coding:utf-8 -*-

# @Time    : 2024/7/26 11:01
# @Author  : wangchongshi
# @Email   : wangchongshi.wcs@antgroup.com
# @FileName: planner_dto.py
from typing import Optional, List

from pydantic import BaseModel, Field


class PlannerDTO(BaseModel):
    id: str = Field(description="ID")
    nickname: Optional[str] = Field(description="planner nickname", default="")
    members: Optional[list] = Field(description="planner members", default=[])
    workflow_id: Optional[str] = None



# !/usr/bin/env python3
# -*- coding:utf-8 -*-

# @Time    : 2024/8/13 16:25
# @Author  : fanen.lhy
# @Email   : fanen.lhy@antgroup.com
# @FileName: nlu_rag_router.py

import json
from typing import List, Tuple, Optional

from agentuniverse.agent.agent_manager import AgentManager
from agentuniverse.agent.action.knowledge.store.store_manager import StoreManager
from agentuniverse.agent.action.knowledge.rag_router.rag_router import \
    RagRouter
from agentuniverse.agent.action.knowledge.store.query import Query
from agentuniverse.base.config.component_configer.component_configer import \
    ComponentConfiger


class NluRagRouter(RagRouter):
    llm: Optional[dict] = None
    agent_name: str = 'nlu_rag_route_agent'
    store_amount: int = 1

    def _rag_route(self, query: Query, store_list: List[str]) \
            -> List[Tuple[Query, str]]:

        agent = AgentManager().get_instance_obj(self.agent_name)
        if self.llm:
            agent.agent_model.profile['llm_model'] = self.llm
        store_info = {}
        for _store in store_list:
            store_info[_store] = StoreManager().get_instance_obj(_store).description
        agent_result = agent.run(
            query=query.query_str,
            store_info=json.dumps(store_info),
            store_amount=self.store_amount
        )
        target_store_list = agent_result.output.split(",")
        if len(target_store_list) > self.store_amount:
            target_store_list = target_store_list[:self.store_amount]
        filtered_store_list = [_store for _store in target_store_list if _store in store_list]

        return [(query, store) for store in filtered_store_list]

    def _initialize_by_component_configer(self,
                                          rag_router_config: ComponentConfiger) -> 'RagRouter':
        super()._initialize_by_component_configer(rag_router_config)
        if hasattr(rag_router_config, "llm"):
            self.llm = rag_router_config.llm
        if hasattr(rag_router_config, "agent_name"):
            self.agent_name = rag_router_config.agent_name
        if hasattr(rag_router_config, "store_amount"):
            self.store_amount = rag_router_config.store_amount






# !/usr/bin/env python3
# -*- coding:utf-8 -*-

# @Time    : 2024/3/22 15:52
# @Author  : wangchongshi
# @Email   : wangchongshi.wcs@antgroup.com
# @FileName: query.py

from PIL.Image import Image
from typing import Optional, List, Set

from pydantic import BaseModel, Field


class Query(BaseModel):
    """The basic class for the knowledge query.

    Attributes:
        query_str (Optional[str]): Origin query str.
        query_text_bundles (Optional[List[str]]): The query string list.
        query_image_bundles (Optional[List[Image]]): The query image list.
        keywords: (Optional[Set[str]]): Keywords used to search with inverted index.
        embeddings (List[List[float]]): A list of embedded queries.
        ext_info (dict): extra information used in query.
    """
    class Config:
        arbitrary_types_allowed = True

    query_str: Optional[str] = None
    query_text_bundles: Optional[List[str]] = Field(default_factory=list)
    query_image_bundles: Optional[List[Image]] = Field(default_factory=list)
    keywords: Optional[Set[str]] = Field(default_factory=set)
    embeddings: List[List[float]] = Field(default_factory=list)
    ext_info: dict = {}

    similarity_top_k: Optional[int] = None



# !/usr/bin/env python3
# -*- coding:utf-8 -*-

# @Time    : 2024/8/28 14:29
# @Author  : sunshinesmilelk
# @Email   : ximo.lk@antgroup.com
# @FileName: ssrf_proxy.py
import os
import httpx

SSRF_PROXY_ALL_URL = os.getenv('SSRF_PROXY_ALL_URL', '')
SSRF_PROXY_HTTP_URL = os.getenv('SSRF_PROXY_HTTP_URL', '')
SSRF_PROXY_HTTPS_URL = os.getenv('SSRF_PROXY_HTTPS_URL', '')

proxies = {
    'http://': SSRF_PROXY_HTTP_URL,
    'https://': SSRF_PROXY_HTTPS_URL
} if SSRF_PROXY_HTTP_URL and SSRF_PROXY_HTTPS_URL else None


def make_request(method, url, **kwargs):
    if SSRF_PROXY_ALL_URL:
        return httpx.request(method=method, url=url, proxy=SSRF_PROXY_ALL_URL, timeout=20, **kwargs)
    elif proxies:
        return httpx.request(method=method, url=url, proxies=proxies, timeout=20, **kwargs)
    else:
        return httpx.request(method=method, url=url, timeout=20, **kwargs)


def get(url, **kwargs):
    return make_request('GET', url, **kwargs)


def post(url, **kwargs):
    return make_request('POST', url, **kwargs)


def put(url, **kwargs):
    return make_request('PUT', url, **kwargs)


def patch(url, **kwargs):
    return make_request('PATCH', url, **kwargs)


def delete(url, **kwargs):
    return make_request('DELETE', url, **kwargs)


def head(url, **kwargs):
    return make_request('HEAD', url, **kwargs)



# !/usr/bin/env python3
# -*- coding:utf-8 -*-

# @Time    : 2024/7/23 17:49
# @Author  : wangchongshi
# @Email   : wangchongshi.wcs@antgroup.com
# @FileName: agentuniverse_product.py
import sys

from agentuniverse.base.agentuniverse import AgentUniverse
from agentuniverse.base.component.application_component_manager import ApplicationComponentManager
from agentuniverse.base.component.component_configer_util import ComponentConfigerUtil
from agentuniverse.base.component.component_enum import ComponentEnum
from agentuniverse.base.config.application_configer.app_configer import AppConfiger
from agentuniverse.base.config.application_configer.application_config_manager import ApplicationConfigManager
from agentuniverse.base.config.component_configer.component_configer import ComponentConfiger
from agentuniverse.base.config.config_type_enum import ConfigTypeEnum
from agentuniverse.base.config.configer import Configer
from agentuniverse.base.util.system_util import get_project_root_path
from agentuniverse.base.annotation.singleton import singleton
from agentuniverse.database.sqldb_wrapper_manager import SQLDBWrapperManager
from agentuniverse_product.base.product import Product
from agentuniverse_product.base.product_configer import ProductConfiger
from agentuniverse_product.base.product_manager import ProductManager
from agentuniverse_product.dal.message_library import MESSAGE_TABLE_NAME, MessageORM
from agentuniverse_product.dal.session_library import SESSION_TABLE_NAME, SessionORM


@singleton
class AgentUniverseProduct(object):
    """Initialize the agentUniverse product."""

    def __init__(self):
        self.__application_container = ApplicationComponentManager()
        self.__config_container: ApplicationConfigManager = ApplicationConfigManager()
        self.__system_default_product_package = []

    def start(self, config_path: str = None):
        """Start the agentUniverse product."""
        # get default config path
        project_root_path = get_project_root_path()
        sys.path.append(str(project_root_path.parent))
        self._add_to_sys_path(project_root_path, ['platform', 'app'])
        if not config_path:
            config_path = project_root_path / 'config' / 'config.toml'
            config_path = str(config_path)

        # load the configuration file
        configer = Configer(path=config_path).load()
        app_configer = AppConfiger().load_by_configer(configer)
        self.__config_container.app_configer = app_configer

        # init product tables (session and message)
        self.__init_product_tables()

        # scan and register the product components
        self.__scan_and_register_product(self.__config_container.app_configer)

        # start the product ui server
        config = configer.value.get('MAGENT_UI', {})
        try:
            from magent_ui import launch
        except ImportError as e:
            print(e)
            raise ImportError(
                "Could not start product server provided by magent-ui."
                " Please install it with `pip install magent-ui ruamel.yaml`."
            )
        launch(**config)

    def __init_product_tables(self):
        """Initialize the product tables including session and message tables."""
        system_sqldb_wrapper = SQLDBWrapperManager().get_instance_obj('__system_db__')
        if system_sqldb_wrapper is None:
            raise Exception('system db has not been initialized.')
        with system_sqldb_wrapper.sql_database._engine.connect() as conn:
            # init session db
            if not conn.dialect.has_table(conn, SESSION_TABLE_NAME):
                SessionORM.metadata.create_all(
                    system_sqldb_wrapper.sql_database._engine)
            # init message db
            if not conn.dialect.has_table(conn, MESSAGE_TABLE_NAME):
                MessageORM.metadata.create_all(
                    system_sqldb_wrapper.sql_database._engine)

    def __scan_and_register_product(self, app_configer: AppConfiger):
        """Scan the product component directory and register the product components.

        Args:
            app_configer(AppConfiger): the AppConfiger object
        """
        core_product_package_list = ((app_configer.core_product_package_list or app_configer.core_default_package_list)
                                     + self.__system_default_product_package)
        if core_product_package_list is None:
            return
        product_configer_list = AgentUniverse().scan(core_product_package_list, ConfigTypeEnum.YAML,
                                                     ComponentEnum.PRODUCT)
        self.__register(product_configer_list)

    def __register(self, product_configer_list: list[ComponentConfiger]):
        """Register the product components.

        Args:
            product_configer_list(list): the product component configer list
        """
        for product_configer in product_configer_list:
            configer_instance: ProductConfiger = ProductConfiger(
            ).load_by_configer(product_configer.configer)
            component_clz = ComponentConfigerUtil.get_component_object_clz_by_component_configer(
                configer_instance)
            product_instance: Product = component_clz(
            ).initialize_by_component_configer(configer_instance)
            if product_instance is None:
                continue
            product_instance.component_config_path = product_configer.configer.path
            ProductManager().register(product_instance.get_instance_code(), product_instance)

    def _add_to_sys_path(self, root_path, sub_dirs):
        for sub_dir in sub_dirs:
            app_path = root_path / sub_dir
            if app_path.exists():
                sys.path.append(str(app_path))



# !/usr/bin/env python3
# -*- coding:utf-8 -*-

# @Time    : 2024/3/27 11:37
# @Author  : wangchongshi
# @Email   : wangchongshi.wcs@antgroup.com
# @FileName: memory_util.py
from typing import List

from langchain_core.chat_history import BaseChatMessageHistory

from agentuniverse.agent.memory.enum import ChatMessageEnum
from agentuniverse.agent.memory.message import Message
from agentuniverse.llm.llm import LLM
from agentuniverse.llm.llm_manager import LLMManager


def generate_messages(memories: list) -> List[Message]:
    """ Generate a list of messages from the given memories

    Args:
        memories(list): List of memory objects, which can be of type str, dict or Message.

    Returns:
        List[Message]: List of messages
    """
    messages = []
    for m in memories:
        if isinstance(m, Message):
            messages.append(m)
        elif isinstance(m, dict):
            messages.append(Message.from_dict(m))
        elif isinstance(m, str):
            message: Message = Message(content=m, metadata={})
            messages.append(message)
    return messages


def generate_memories(chat_messages: BaseChatMessageHistory) -> list:
    return [
        {"content": message.content, "type": 'ai' if message.type == 'AIMessageChunk' else message.type}
        for message in chat_messages.messages
    ] if chat_messages.messages else []


def get_memory_string(messages: List[Message]) -> str:
    """Convert the given messages to a string.

    Args:
        messages(List[Message]): The list of messages.

    Returns:
        str: The string representation of the messages.
    """

    string_messages = []
    for m in messages:
        if m.type == ChatMessageEnum.SYSTEM.value:
            role = 'System'
        elif m.type == ChatMessageEnum.HUMAN.value:
            role = 'Human'
        elif m.type == ChatMessageEnum.AI.value:
            role = "AI"
        else:
            role = ""
        m_str = ""
        if role:
            m_str += f"Message role: {role} "
        if m.source:
            m_str += f"Message source: {m.source} "
        m_str += f"Message content: \n {m.content} "
        string_messages.append(m_str)
    return "\n\n".join(string_messages)


def get_memory_tokens(memories: List[Message], llm_name: str = None) -> int:
    """Get the number of tokens in the given memories.

    Args:
        memories(List[Message]): The list of messages.
        llm_name(str): The name of the LLM to use for token counting.

    Returns:
        int: The number of tokens in the given memories.
    """
    memory_str = get_memory_string(memories)
    llm_instance: LLM = LLMManager().get_instance_obj(llm_name)
    return llm_instance.get_num_tokens(memory_str) if llm_instance else len(memory_str)



# !/usr/bin/env python3
# -*- coding:utf-8 -*-

# @Time    : 2024/3/22 15:50
# @Author  : wangchongshi
# @Email   : wangchongshi.wcs@antgroup.com
# @FileName: store.py
from typing import Any, List, Optional

from agentuniverse.base.component.component_base import ComponentEnum
from agentuniverse.base.component.component_base import ComponentBase
from agentuniverse.base.config.component_configer.component_configer import ComponentConfiger
from agentuniverse.agent.action.knowledge.store.document import Document
from agentuniverse.agent.action.knowledge.store.query import Query
from agentuniverse.agent_serve.web.post_fork_queue import add_post_fork


class Store(ComponentBase):
    """The basic class for the knowledge store.

    Store of the knowledge, store class is used to store knowledge
    and provide retrieval capabilities,
    vector storage, such as ChromaDB store, or non-vector storage, such as Redis Store.
    """
    component_type: ComponentEnum = ComponentEnum.STORE
    name: Optional[str] = None
    description: Optional[str] = None
    client: Any = None
    async_client: Any = None

    class Config:
        """Configuration for this pydantic object."""
        arbitrary_types_allowed = True

    def _new_client(self) -> Any:
        """Initialize the client，like database connection or local file create."""
        pass

    def _new_async_client(self) -> Any:
        """Initialize the async client."""
        pass

    def _initialize_by_component_configer(self,
                                         store_configer: ComponentConfiger) \
            -> 'Store':
        """Initialize the store by the ComponentConfiger object.

        Args:
            store_configer(ComponentConfiger): A configer contains store
            basic info. The client will be created in post fork due to
            conflicts that may occur in a multithreaded scenario.
        Returns:
            Store: A store instance.
        """
        if store_configer.name:
            self.name = store_configer.name
        if store_configer.description:
            self.description = store_configer.description
        add_post_fork(self._new_client)
        add_post_fork(self._new_async_client)
        return self

    def query(self, query: Query, **kwargs) -> List[Document]:
        """Query documents."""
        raise NotImplementedError

    async def async_query(self, query: Query, **kwargs) -> List[Document]:
        """Asynchronously query documents."""
        raise NotImplementedError

    def insert_document(self, documents: List[Document], **kwargs):
        """Insert documents into the store."""
        raise NotImplementedError

    async def async_insert_document(self, documents: List[Document], **kwargs):
        """Asynchronously insert documents into the store."""
        raise NotImplementedError

    def delete_document(self, document_id: str, **kwargs):
        """Delete the specific document by the document id."""
        raise NotImplementedError

    async def async_delete_document(self, document_id: str, **kwargs):
        """Asynchronously delete the specific document by the document id."""
        raise NotImplementedError

    def upsert_document(self, documents: List[Document], **kwargs):
        """Upsert document into the store."""
        raise NotImplementedError

    async def async_upsert_document(self, documents: List[Document], **kwargs):
        """Asynchronously upsert documents into the store."""
        raise NotImplementedError

    def update_document(self, documents: List[Document], **kwargs):
        """Update document into the store."""
        raise NotImplementedError

    async def async_update_document(self, documents: List[Document], **kwargs):
        """Asynchronously update documents into the store."""
        raise NotImplementedError



# !/usr/bin/env python3
# -*- coding:utf-8 -*-

# @Time    : 2024/3/13 12:01
# @Author  : jerry.zzw 
# @Email   : jerry.zzw@antgroup.com
# @FileName: agent_configer.py
from typing import Optional
from agentuniverse.base.config.component_configer.component_configer import ComponentConfiger
from agentuniverse.base.config.configer import Configer


class AgentConfiger(ComponentConfiger):
    """The AgentConfiger class, which is used to load and manage the Agent configuration."""

    def __init__(self, configer: Optional[Configer] = None):
        """Initialize the AgentConfiger."""
        super().__init__(configer)
        self.__info: Optional[dict] = dict()
        self.__profile: Optional[dict] = dict()
        self.__plan: Optional[dict] = dict()
        self.__memory: Optional[dict] = dict()
        self.__action: Optional[dict] = dict()

    @property
    def memory(self) -> Optional[dict]:
        """Return the name of the Agent."""
        return self.__memory

    @property
    def action(self) -> Optional[dict]:
        """Return the name of the Agent."""
        return self.__action

    @property
    def profile(self) -> Optional[dict]:
        """Return the name of the Agent."""
        return self.__profile

    @property
    def plan(self) -> Optional[dict]:
        """Return the description of the Agent."""
        return self.__plan

    @property
    def info(self) -> Optional[dict]:
        """Return the name of the Agent."""
        return self.__info

    def load(self) -> 'AgentConfiger':
        """Load the configuration by the Configer object.
        Returns:
            AgentConfiger: the AgentConfiger object
        """
        return self.load_by_configer(self.configer)

    def load_by_configer(self, configer: Configer) -> 'AgentConfiger':
        """Load the configuration by the Configer object.
        Args:
            configer(Configer): the Configer object
        Returns:
            AgentConfiger: the AgentConfiger object
        """
        super().load_by_configer(configer)

        try:
            configer_value: dict = configer.value
            self.__info = configer_value.get('info') or self.__info
            self.__profile = configer_value.get('profile') or self.__profile
            self.__plan = configer_value.get('plan') or self.__plan
            self.__memory = configer_value.get('memory') or self.__memory
            self.__action = configer_value.get('action') or self.__action
        except Exception as e:
            raise Exception(f"Failed to parse the Agent configuration: {e}")
        return self



from .service import Service
from .service_manager import ServiceManager


class ServiceNotFoundError(Exception):
    """An exception when service code is not in service manager."""
    def __init__(self, service_code: str):
        super().__init__(f"Service {service_code} not found.")
        self.service_code = service_code


class ServiceInstance(object):
    """A service wrapper class, which should be directly called in project
    instead of Service class."""

    def __init__(self, service_code: str):
        """Initialize a service instance. Raise an ServiceNotFoundError when
        service code can't be found by servie manager.

        Args:
            service_code (`str`):
                Unique code of the service.
        """
        self.__service_code = service_code
        service_manager: ServiceManager = ServiceManager()
        self.__service: Service = service_manager.get_instance_obj(
            service_code
        )
        if self.__service is None:
            raise ServiceNotFoundError(service_code)

    def run(self, **kwargs) -> str:
        """Call the service run."""
        return self.__service.run(**kwargs)



# !/usr/bin/env python3
# -*- coding:utf-8 -*-

# @Time    : 2024/10/17 20:37
# @Author  : wangchongshi
# @Email   : wangchongshi.wcs@antgroup.com
# @FileName: expressing_agent_template.py
from queue import Queue

from agentuniverse.agent.input_object import InputObject
from agentuniverse.agent.template.agent_template import AgentTemplate
from agentuniverse.base.config.component_configer.configers.agent_configer import AgentConfiger
from agentuniverse.base.util.common_util import stream_output
from agentuniverse.base.util.logging.logging_util import LOGGER


class ExpressingAgentTemplate(AgentTemplate):

    def input_keys(self) -> list[str]:
        return ['input', 'executing_result']

    def output_keys(self) -> list[str]:
        return ['output']

    def parse_input(self, input_object: InputObject, agent_input: dict) -> dict:
        agent_input['input'] = input_object.get_data('input')
        agent_input['background'] = self.build_execution_context(input_object)
        agent_input['expert_framework'] = input_object.get_data('expert_framework', {}).get('expressing')
        return agent_input

    def parse_result(self, agent_result: dict) -> dict:
        final_result = dict()
        final_result['output'] = agent_result['output']
        # add expressing agent log info.
        logger_info = f"\nExpressing agent execution result is :\n"
        logger_info += f"{final_result.get('output')}"
        LOGGER.info(logger_info)

        return final_result

    def build_execution_context(self, input_object: InputObject) -> str:
        executing_result = input_object.get_data('executing_result').get_data('executing_result', [])
        execution_context_list = [
            f"question:{execution.get('input')}\nanswer:{execution.get('output')}"
            for execution in executing_result
        ]
        return '\n\n'.join(execution_context_list)

    def initialize_by_component_configer(self, component_configer: AgentConfiger) -> 'ExpressingAgentTemplate':
        super().initialize_by_component_configer(component_configer)
        self.prompt_version = self.agent_model.profile.get('prompt_version', 'default_expressing_agent.cn')
        self.validate_required_params()
        return self

    def validate_required_params(self):
        if not self.llm_name:
            raise ValueError(f'llm_name of the agent {self.agent_model.info.get("name")}'
                             f' is not set, please go to the agent profile configuration'
                             ' and set the `name` attribute in the `llm_model`.')

    def add_output_stream(self, output_stream: Queue, agent_output: str) -> None:
        if not output_stream:
            return
        # add expressing agent final result into the stream output.
        stream_output(output_stream,
                      {"data": {
                          'output': agent_output,
                          "agent_info": self.agent_model.info
                      }, "type": "expressing"})



# !/usr/bin/env python3
# -*- coding:utf-8 -*-

# @Time    : 2024/3/13 12:01
# @Author  : jerry.zzw 
# @Email   : jerry.zzw@antgroup.com
# @FileName: llm_configer.py
from typing import Optional
from agentuniverse.base.config.component_configer.component_configer import ComponentConfiger
from agentuniverse.base.config.configer import Configer


class LLMConfiger(ComponentConfiger):
    """The LLMConfiger class, which is used to load and manage the LLM configuration."""

    def __init__(self, configer: Optional[Configer] = None):
        """Initialize the LLMConfiger."""
        super().__init__(configer)
        self.__name: Optional[str] = None
        self.__description: Optional[str] = None
        self.__model_name: Optional[str] = None
        self.__temperature: Optional[float] = None
        self.__request_timeout: Optional[int] = None
        self.__max_tokens: Optional[int] = None
        self.__max_retries: Optional[int] = None
        self.__streaming: Optional[bool] = None
        self.__ext_info: Optional[dict] = None
        self.__max_context_length: Optional[int] = None
        self.__tracing: Optional[bool] = None

    @property
    def name(self) -> Optional[str]:
        """Return the name of the LLM."""
        return self.__name

    @property
    def description(self) -> Optional[str]:
        """Return the description of the LLM."""
        return self.__description

    @property
    def model_name(self) -> Optional[str]:
        return self.__model_name

    @property
    def temperature(self) -> Optional[float]:
        """Return the temperature of the LLM."""
        return self.__temperature

    @property
    def request_timeout(self) -> Optional[int]:
        return self.__request_timeout

    @property
    def max_tokens(self) -> Optional[int]:
        return self.__max_tokens

    @property
    def max_retries(self) -> Optional[int]:
        return self.__max_retries

    @property
    def streaming(self) -> Optional[bool]:
        return self.__streaming

    @property
    def ext_info(self) -> Optional[dict]:
        return self.__ext_info

    @property
    def max_content_length(self) -> Optional[int]:
        return self.__max_context_length

    @property
    def tracing(self) -> Optional[bool]:
        return self.__tracing

    def load(self) -> 'LLMConfiger':
        """Load the configuration by the Configer object.
        Returns:
            LLMConfiger: the LLMConfiger object
        """
        return self.load_by_configer(self.__configer)

    def load_by_configer(self, configer: Configer) -> 'LLMConfiger':
        """Load the configuration by the Configer object.
        Args:
            configer(Configer): the Configer object
        Returns:
            LLMConfiger: the LLMConfiger object
        """
        super().load_by_configer(configer)

        try:
            self.__name = configer.value.get('name')
            self.__description = configer.value.get('description')
            self.__model_name = configer.value.get('model_name')
            self.__temperature = configer.value.get('temperature')
            self.__request_timeout = configer.value.get('request_timeout')
            self.__max_tokens = configer.value.get('max_tokens')
            self.__max_retries = configer.value.get('max_retries')
            self.__streaming = configer.value.get('streaming')
            self.__ext_info = configer.value.get('ext_info')
            self.__max_context_length = configer.value.get('max_context_length')
            self.__tracing = configer.value.get('tracing')
        except Exception as e:
            raise Exception(f"Failed to parse the LLM configuration: {e}")
        return self



# !/usr/bin/env python3
# -*- coding:utf-8 -*-

# @Time    : 2024/5/22 17:07
# @Author  : weizjajj 
# @Email   : weizhongjie.wzj@antgroup.com
# @FileName: claude_langchain_instance.py

import warnings
from typing import List, Optional, Any

from langchain_anthropic import ChatAnthropic
from langchain_anthropic.chat_models import _tools_in_params
from langchain_core.callbacks import CallbackManagerForLLMRun, AsyncCallbackManagerForLLMRun
from langchain_core.language_models.chat_models import generate_from_stream, agenerate_from_stream
from langchain_core.messages import BaseMessage, SystemMessage, HumanMessage
from langchain_core.outputs import ChatResult

from agentuniverse.llm.llm import LLM


class ClaudeLangChainInstance(ChatAnthropic):
    """
    This class wraps the Claude API into a LangChain API.
    """
    llm: LLM = None

    def __init__(self, llm: LLM):
        init_params = {}
        init_params['model'] = llm.model_name if llm.model_name else 'Claude-Instant-V1.3'
        init_params['temperature'] = llm.temperature if llm.temperature else 0.7
        init_params['default_request_timeout'] = llm.request_timeout
        init_params['streaming'] = llm.streaming if llm.streaming else False
        init_params['anthropic_api_key'] = llm.anthropic_api_key if llm.anthropic_api_key else 'blank'
        init_params['max_tokens'] = llm.max_tokens
        init_params['max_retries'] = llm.max_retries if llm.max_retries else 2
        init_params['anthropic_api_url'] = llm.anthropic_api_url
        init_params['streaming'] = llm.streaming
        init_params['llm'] = llm
        super().__init__(**init_params)

    def _generate(
            self,
            messages: List[BaseMessage],
            stop: Optional[List[str]] = None,
            run_manager: Optional[CallbackManagerForLLMRun] = None,
            **kwargs: Any,
    ) -> ChatResult:
        if len(messages) > 1 and isinstance(messages[1], SystemMessage):
            messages[1] = HumanMessage(content=messages[1].content)
        params = self._format_params(messages=messages, stop=stop, **kwargs)
        if self.streaming:
            if _tools_in_params(params):
                warnings.warn(
                    "stream: Tool use is not yet supported in streaming mode."
                )
            else:
                stream_iter = self._stream(
                    messages, stop=stop, run_manager=run_manager, **kwargs
                )
                return generate_from_stream(stream_iter)
        if _tools_in_params(params):
            data = self._client.beta.tools.messages.create(**params)
        else:
            data = self.llm.call(**params).raw
        return self._format_output(data, **kwargs)

    async def _agenerate(
            self,
            messages: List[BaseMessage],
            stop: Optional[List[str]] = None,
            run_manager: Optional[AsyncCallbackManagerForLLMRun] = None,
            **kwargs: Any,
    ) -> ChatResult:
        if len(messages) > 1 and isinstance(messages[1], SystemMessage):
            messages[1] = HumanMessage(content=messages[1].content)
        params = self._format_params(messages=messages, stop=stop, **kwargs)
        if self.streaming:
            if _tools_in_params(params):
                warnings.warn(
                    "stream: Tool use is not yet supported in streaming mode."
                )
            else:
                stream_iter = self._astream(
                    messages, stop=stop, run_manager=run_manager, **kwargs
                )
                return await agenerate_from_stream(stream_iter)
        if _tools_in_params(params):
            data = await self._async_client.beta.tools.messages.create(**params)
        else:
            data = await self.llm.acall(**params)
            data = data.raw
        return self._format_output(data, **kwargs)



# !/usr/bin/env python3
# -*- coding:utf-8 -*-

# @Time    : 2024/7/25 19:34
# @Author  : wangchongshi
# @Email   : wangchongshi.wcs@antgroup.com
# @FileName: message_library.py
import datetime

from sqlalchemy import JSON, Integer, String, DateTime, Column, Index, asc
from sqlalchemy.orm import declarative_base

from agentuniverse.base.annotation.singleton import singleton
from agentuniverse.database.sqldb_wrapper_manager import SQLDBWrapperManager
from agentuniverse_product.dal.model.message_do import MessageDO

MESSAGE_TABLE_NAME = 'message'
Base = declarative_base()


class MessageORM(Base):
    """Sqlalchemy orm Model for message table."""
    __tablename__ = MESSAGE_TABLE_NAME
    id = Column(Integer, primary_key=True, autoincrement=True)
    session_id = Column(String(50), nullable=False)
    content = Column(JSON)
    ext_info = Column(JSON)
    gmt_created = Column(DateTime, default=datetime.datetime.now)
    gmt_modified = Column(DateTime, default=datetime.datetime.now,
                          onupdate=datetime.datetime.now)
    __table_args__ = (
        Index('message_ix_session_id', 'session_id'),
    )


@singleton
class MessageLibrary:
    @staticmethod
    def get_db_session():
        """Get the database session."""
        system_sqldb_wrapper = SQLDBWrapperManager().get_instance_obj('__system_db__')
        return system_sqldb_wrapper.get_session()()

    def add_message(self, message_do: MessageDO) -> int:
        """Add a message to the database."""
        with self.get_db_session() as db_session:
            message_orm = MessageORM(**message_do.model_dump())
            db_session.add(message_orm)
            db_session.commit()
            return message_orm.id

    def delete_messages(self, session_id: str):
        """Delete messages from the database using the provided `session_id`."""
        with self.get_db_session() as db_session:
            message_orm_list = db_session.query(MessageORM).filter(
                MessageORM.session_id == session_id)
            if message_orm_list:
                for message_orm in message_orm_list:
                    db_session.delete(message_orm)
                db_session.commit()

    def get_messages(self, session_id: str) -> list[MessageDO]:
        """Get messages from the database using the provided `session_id`."""
        with self.get_db_session() as db_session:
            message_orm_list = db_session.query(MessageORM).filter(
                MessageORM.session_id == session_id).order_by(asc(MessageORM.gmt_modified)).all()
            res = []
            if message_orm_list:
                for message_orm in message_orm_list:
                    res.append(self.__message_orm_to_do(message_orm))
            return res

    @staticmethod
    def __message_orm_to_do(message_orm: MessageORM) -> MessageDO:
        """Convert a MessageORM object to a MessageDO object."""
        message_do = MessageDO(
            session_id="",
            content='',
            ext_info=dict(),
        )
        for column in message_orm.__table__.columns:
            setattr(message_do, column.name,
                    getattr(message_orm, column.name))
        return message_do



# !/usr/bin/env python3
# -*- coding:utf-8 -*-

# @Time    : 2024/8/6 15:39
# @Author  : wangchongshi
# @Email   : wangchongshi.wcs@antgroup.com
# @FileName: knowledge_service.py
import os
import shutil

from typing import List

from agentuniverse.base.component.component_enum import ComponentEnum
from agentuniverse_product.base.product import Product
from agentuniverse_product.base.product_manager import ProductManager
from agentuniverse_product.service.model.knowledge_dto import KnowledgeDTO
from agentuniverse.agent.action.knowledge.knowledge_manager import KnowledgeManager
from agentuniverse.agent.action.knowledge.knowledge import Knowledge
from agentuniverse_product.base.util.yaml_util import (
    write_yaml_file,
    delete_yaml_file,
    update_nested_yaml_value,
    read_yaml_file,
)
from agentuniverse_product.service.util.common_util import get_core_path, get_resources_path
from agentuniverse_product.service.util.knowledge_util import (
    register_knowledge,
    unregister_knowledge,
    register_store,
    assemble_knowledge_product_config_data,
    assemble_knowledge_config,
)
from agentuniverse_product.service.util.agent_util import register_product, unregister_product


class KnowledgeService:
    """Knowledge Service for aU-product."""

    @staticmethod
    def get_knowledge_list() -> List[KnowledgeDTO]:
        """Get all knowledge."""
        res = []
        product_list: List[Product] = ProductManager().get_instance_obj_list()
        if len(product_list) < 1:
            return res
        for product in product_list:
            if product.type == ComponentEnum.KNOWLEDGE.value:
                knowledge_dto = KnowledgeDTO(nickname=product.nickname, id=product.id)
                knowledge = product.instance
                knowledge_dto.description = knowledge.description
                res.append(knowledge_dto)
        return res

    @staticmethod
    def create_knowledge(knowledge_dto: KnowledgeDTO) -> str:
        """ Create knowledge with name and desc, then return knowledge ID.

        Args:
            knowledge_dto(KnowledgeDTO): The knowledge DTO containing the knowledge attributes.

        Returns:
            str: The ID of the created knowledge.
        """
        # write product YAML file
        knowledge_id = knowledge_dto.id

        product_file_name = f"{knowledge_id}_product"

        path = get_core_path()
        product_file_path = path / 'product' / 'knowledge' / f"{product_file_name}.yaml" if path\
            else os.path.join("..", "..", "platform", "difizen", "product", "knowledge", f"{product_file_name}.yaml")

        product_config_data = assemble_knowledge_product_config_data(knowledge_dto)

        try:
            write_yaml_file(str(product_file_path), product_config_data)
        except Exception as e:
            raise e

        knowledge_file_path = path / 'knowledge' / f"{knowledge_id}.yaml" if path\
            else os.path.join("..", "..", "intelligence", "agentic", "knowledge", f"{knowledge_id}.yaml")

        knowledge_config = assemble_knowledge_config(knowledge_dto)

        try:
            write_yaml_file(str(knowledge_file_path), knowledge_config)
        except Exception as e:
            raise e

        register_knowledge(str(knowledge_file_path))
        register_product(str(product_file_path))

        return knowledge_id

    @staticmethod
    def update_knowledge(knowledge_dto: KnowledgeDTO) -> str:
        """Update knowledge attributes.

        Args:
            knowledge_dto (KnowledgeDTO): The knowledge DTO containing the knowledge attributes.

        Returns:
            str: The ID of the updated knowledge.
        """

        if knowledge_dto.id is None:
            raise ValueError("knowledge_id is None.")

        knowledge: Knowledge = KnowledgeManager().get_instance_obj(component_instance_name=knowledge_dto.id)
        if knowledge is None:
            raise ValueError("The knowledge instance corresponding to the knowledge_id cannot be found.")

        product: Product = ProductManager().get_instance_obj(component_instance_name=knowledge_dto.id)
        if product is None:
            raise ValueError("The knowledge product instance corresponding to the knowledge_id cannot be found.")

        knowledge_update_config = {"description": knowledge_dto.description}
        product_update_config = {"nickname": knowledge_dto.nickname}
        # Update the knowledge attributes if they are provided
        knowledge_updated_fields = {key: value for key, value in knowledge_update_config.items() if value is not None}
        product_updated_fields = {key: value for key, value in product_update_config.items() if value is not None}

        # set knowledge attributes
        for key, value in knowledge_updated_fields.items():
            setattr(knowledge, key, value)
        for key, value in product_updated_fields.items():
            setattr(product, key, value)

        update_nested_yaml_value(knowledge.component_config_path, knowledge_updated_fields)
        update_nested_yaml_value(product.component_config_path, product_updated_fields)

        return knowledge_dto.id

    @staticmethod
    def delete_knowledge(knowledge_id: str) -> bool:
        """Delete knowledge, then return the result.

        Args:
            knowledge_id (str): The ID of the knowledge to be deleted.

        Returns:
            bool: True if the knowledge is deleted successfully, False otherwise.
        """

        knowledge: Knowledge = KnowledgeManager().get_instance_obj(component_instance_name=knowledge_id)
        if knowledge is None:
            raise ValueError("The knowledge instance corresponding to the knowledge_id cannot be found.")

        product: Product = ProductManager().get_instance_obj(component_instance_name=knowledge_id)
        if product is None:
            raise ValueError("The knowledge product instance corresponding to the knowledge_id cannot be found.")

        knowledge_file_path = knowledge.component_config_path
        product_file_path = product.component_config_path

        try:
            unregister_knowledge(knowledge_file_path)
            unregister_product(product_file_path)
            delete_yaml_file(knowledge_file_path)
            delete_yaml_file(product_file_path)
            return True
        except Exception as e:
            raise e

    @staticmethod
    def upload_knowledge_file(knowledge_id: str, file) -> bool:
        """Upload the knowledge file.

        Args:
            knowledge_id (str): The ID of the knowledge.
            file: The file to be uploaded.

        Returns:
            bool: True if the file is uploaded successfully, False otherwise.
        """
        path = get_core_path()

        upload_file_path = get_resources_path()

        try:
            if not os.path.exists(upload_file_path):
                os.makedirs(upload_file_path, exist_ok=True)

            file_location = os.path.join(upload_file_path, file.filename)

            with open(file_location, "wb") as buffer:
                shutil.copyfileobj(file.file, buffer)

        except Exception as e:
            print(f"upload file failed: {e}")
            raise e

        knowledge_store_file_path = path / 'store' / f"{knowledge_id}_store.yaml" if path\
            else str(os.path.join("..", "..", "intelligence",
                                  "agentic", "knowledge", "store", f"{knowledge_id}_store.yaml"))
        knowledge_store_name = f"{knowledge_id}_chroma_store"

        knowledge = KnowledgeManager().get_instance_obj(component_instance_name=knowledge_id)

        try:
            # If the store YAML file does not exist, create and register the store, and update the knowledge YAML.
            if not os.path.isfile(knowledge_store_file_path):
                knowledge_store_config = {
                    "name": knowledge_store_name,
                    "description": knowledge_store_name,
                    "persist_path": f"../../intelligence/db/{knowledge_store_name}.db",
                    "metadata": {
                        "type": "STORE",
                        "module": "agentuniverse.agent.action.knowledge.store.chroma_store",
                        "class": "ChromaStore",
                    },
                }
                write_yaml_file(str(knowledge_store_file_path), knowledge_store_config)
                register_store(str(knowledge_store_file_path))

                knowledge_path = knowledge.component_config_path

                knowledge_config = read_yaml_file(knowledge_path)
                knowledge_store = knowledge_config.get("stores", [])

                update_nested_yaml_value(
                    knowledge.component_config_path, {"stores": [*knowledge_store, knowledge_store_name]}
                )

        except Exception as e:
            raise e
        stores = knowledge.stores
        stores.append(knowledge_store_name)
        knowledge.insert_knowledge(source_path=file_location, stores=[knowledge_store_name])

        return True



# !/usr/bin/env python3
# -*- coding:utf-8 -*-

# @Time    : 2024/8/29 10:16
# @Author  : wangchongshi
# @Email   : wangchongshi.wcs@antgroup.com
# @FileName: common_util.py
from agentuniverse.base.component.component_base import ComponentBase
from agentuniverse.base.component.component_configer_util import ComponentConfigerUtil
from agentuniverse.base.component.component_enum import ComponentEnum


def is_component_id_unique(component_id: str, component_type: str) -> bool:
    """Check if the component ID is unique.

    Args:
        component_id (str): The component ID to check.
        component_type (str): The component type.

    Returns:
        bool: True if the component ID is unique, False otherwise.
    """
    if component_id is None or component_type is None:
        return True
    component_enum = ComponentEnum.from_value(component_type.upper())
    component_manager_clz = ComponentConfigerUtil.get_component_manager_clz_by_type(component_enum)
    instance: ComponentBase = component_manager_clz().get_instance_obj(component_id)
    return instance is None



# !/usr/bin/env python3
# -*- coding:utf-8 -*-

# @Time    : 2024/8/5 14:37
# @Author  : fanen.lhy
# @Email   : fanen.lhy@antgroup.com
# @FileName: text_splitter.py
from typing import List, Optional
from langchain.text_splitter import CharacterTextSplitter as Splitter

from agentuniverse.agent.action.knowledge.doc_processor.doc_processor import \
    DocProcessor
from agentuniverse.agent.action.knowledge.store.document import Document
from agentuniverse.agent.action.knowledge.store.query import Query
from agentuniverse.base.config.component_configer.component_configer import \
    ComponentConfiger


class CharacterTextSplitter(DocProcessor):
    chunk_size: int = 200
    chunk_overlap: int = 20
    separator: str = "/n/n"
    splitter: Optional[Splitter] = None

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.splitter = Splitter(separator=self.separator,
                                 chunk_size=self.chunk_size,
                                 chunk_overlap=self.chunk_overlap)

    def _process_docs(self, origin_docs: List[Document], query: Query = None) -> \
            List[Document]:
        lc_doc_list = self.splitter.split_documents(Document.as_langchain_list(
            origin_docs
        ))
        return Document.from_langchain_list(lc_doc_list)

    def _initialize_by_component_configer(self,
                                         doc_processor_configer: ComponentConfiger) -> 'DocProcessor':
        super()._initialize_by_component_configer(doc_processor_configer)
        if hasattr(doc_processor_configer, "chunk_size"):
            self.chunk_size = doc_processor_configer.chunk_size
        if hasattr(doc_processor_configer, "chunk_overlap"):
            self.chunk_overlap = doc_processor_configer.chunk_overlap
        if hasattr(doc_processor_configer, "separator"):
            self.separator = doc_processor_configer.separator
        self.splitter = Splitter(separator=self.separator,
                                 chunk_size=self.chunk_size,
                                 chunk_overlap=self.chunk_overlap)
        return self



# !/usr/bin/env python3
# -*- coding:utf-8 -*-

# @Time    : 2024/5/31 21:22
# @Author  : wangchongshi
# @Email   : wangchongshi.wcs@antgroup.com
# @FileName: react_agent.py
from agentuniverse.agent.template.react_agent_template import ReActAgentTemplate


class ReActAgent(ReActAgentTemplate):
    """ReAct Agent class."""



# !/usr/bin/env python3
# -*- coding:utf-8 -*-

# @Time    : 2024/8/30 10:25
# @Author  : wangchongshi
# @Email   : wangchongshi.wcs@antgroup.com
# @FileName: knowledge_util.py
import os

from typing import Dict

from agentuniverse.agent.action.knowledge.knowledge import Knowledge
from agentuniverse.agent.action.knowledge.knowledge_manager import KnowledgeManager
from agentuniverse.agent.action.knowledge.store.store_manager import StoreManager
from agentuniverse.agent_serve.web.post_fork_queue import POST_FORK_QUEUE
from agentuniverse.base.component.component_configer_util import ComponentConfigerUtil
from agentuniverse.base.config.component_configer.component_configer import ComponentConfiger
from agentuniverse.base.config.component_configer.configers.knowledge_configer import KnowledgeConfiger
from agentuniverse.base.config.configer import Configer
from agentuniverse_product.service.model.knowledge_dto import KnowledgeDTO


def assemble_knowledge_product_config_data(knowledge_dto: KnowledgeDTO) -> Dict:
    """Asemble the knowledge product configuration data.

    Args:
        knowledge_dto (KnowledgeDTO): The knowledge DTO.

    Returns:
        Dict: The assembled knowledge product configuration data.
    """
    return {
        'id': knowledge_dto.id,
        'nickname': knowledge_dto.nickname,
        'avatar': knowledge_dto.avatar,
        'type': 'KNOWLEDGE',
        'metadata': {
            'class': 'AgentProduct',
            'module': 'agentuniverse_product.base.agent_product',
            'type': 'PRODUCT'
        }
    }


def assemble_knowledge_config(knowledge_dto: KnowledgeDTO) -> Dict:
    """Asemble the knowledge configuration data.

    Args:
        knowledge_dto (KnowledgeDTO): The knowledge DTO object containing the knowledge parameters.

    Returns:
        Dict: The assembled knowledge configuration data.
    """
    return {
        'name': knowledge_dto.id,
        'description': knowledge_dto.description,
        'stores': [],
        'insert_processors': ["recursive_character_text_splitter"],
        'readers': {
            'pdf': "default_pdf_reader",
            'docx': "default_docx_reader",
            'pptx': "default_pptx_reader",
            'txt': "default_txt_reader",
        },
        'metadata': {
            'type': 'KNOWLEDGE',
            'module': 'agentuniverse.agent.action.knowledge.knowledge',
            'class': 'Knowledge',
        }
    }


def register_knowledge(file_path: str):
    """Register the knowledge instance to the knowledge manager.

    Args:
        file_path (str): The path to the knowledge configuration file.
    """
    absolute_file_path = os.path.abspath(file_path)
    configer = Configer(path=absolute_file_path).load()
    component_configer = ComponentConfiger().load_by_configer(configer)
    knowledge_configer: KnowledgeConfiger = KnowledgeConfiger().load_by_configer(component_configer.configer)
    component_clz = ComponentConfigerUtil.get_component_object_clz_by_component_configer(knowledge_configer)
    component_instance: Knowledge = component_clz().initialize_by_component_configer(knowledge_configer)
    component_instance.component_config_path = component_configer.configer.path
    KnowledgeManager().register(component_instance.get_instance_code(), component_instance)


def unregister_knowledge(file_path: str):
    """Unregister the knowledge instance from the knowledge manager.

    Args:
        file_path (str): The path to the knowledge configuration file.
    """
    absolute_file_path = os.path.abspath(file_path)
    configer = Configer(path=absolute_file_path).load()
    component_configer = ComponentConfiger().load_by_configer(configer)
    knowledge_configer: KnowledgeConfiger = KnowledgeConfiger().load_by_configer(component_configer.configer)
    component_clz = ComponentConfigerUtil.get_component_object_clz_by_component_configer(knowledge_configer)
    component_instance: Knowledge = component_clz().initialize_by_component_configer(knowledge_configer)
    KnowledgeManager().unregister(component_instance.get_instance_code())


def register_store(file_path: str):
    """Register the store instance to the store manager.

    Args:
        file_path (str): The path to the store configuration file.
    """
    absolute_file_path = os.path.abspath(file_path)
    configer = Configer(path=absolute_file_path).load()
    component_configer = ComponentConfiger().load_by_configer(configer)
    component_clz = ComponentConfigerUtil.get_component_object_clz_by_component_configer(component_configer)
    component_instance = component_clz().initialize_by_component_configer(component_configer)
    component_instance.component_config_path = component_configer.configer.path
    StoreManager().register(component_instance.get_instance_code(), component_instance)
    for _func, args, kwargs in POST_FORK_QUEUE[-2:]:
        _func(*args, **kwargs)



# !/usr/bin/env python3
# -*- coding:utf-8 -*-

# @Time    : 2024/4/2 11:43
# @Author  : fanen.lhy
# @Email   : fanen.lhy@antgroup.com
# @FileName: custom_key_configer.py

import os

from agentuniverse.base.annotation.singleton import singleton
from ..configer import Configer


@singleton
class CustomKeyConfiger(Configer):
    """Use to manage user secret key."""
    def __init__(self, config_path: str = None):
        self._Configer__value = {}
        super().__init__(config_path)
        if config_path:
            try:
                self.load()
            except FileNotFoundError as e:
                print(f"Custom key file {config_path} read error, "
                      f"skip load custom key.")
        if self._Configer__value.get("KEY_LIST"):
            for key, value in self._Configer__value.get("KEY_LIST").items():
                os.environ[key] = str(value)


# !/usr/bin/env python3
# -*- coding:utf-8 -*-

# @Time    : 2024/7/25 19:34
# @Author  : wangchongshi
# @Email   : wangchongshi.wcs@antgroup.com
# @FileName: session_library.py
import datetime

from sqlalchemy import JSON, Integer, String, DateTime, Column, Index, desc
from sqlalchemy.orm import declarative_base

from agentuniverse.base.annotation.singleton import singleton
from agentuniverse.database.sqldb_wrapper_manager import SQLDBWrapperManager
from agentuniverse_product.dal.model.session_do import SessionDO
from agentuniverse.base.util.logging.logging_util import LOGGER

SESSION_TABLE_NAME = 'session'
Base = declarative_base()


class SessionORM(Base):
    """Sqlalchemy orm Model for session table."""
    __tablename__ = SESSION_TABLE_NAME
    id = Column(Integer, primary_key=True, autoincrement=True)
    session_id = Column(String(50), nullable=False)
    agent_id = Column(String(50), nullable=False)
    ext_info = Column(JSON)
    gmt_created = Column(DateTime, default=datetime.datetime.now)
    gmt_modified = Column(DateTime, default=datetime.datetime.now,
                          onupdate=datetime.datetime.now)
    __table_args__ = (
        Index('session_ix_session_id', 'session_id'),
        Index('session_ix_agent_id', 'agent_id'),
    )


@singleton
class SessionLibrary:

    @staticmethod
    def get_db_session():
        """Get the database session."""
        system_sqldb_wrapper = SQLDBWrapperManager().get_instance_obj('__system_db__')
        return system_sqldb_wrapper.get_session()()

    def add_session(self, session_do: SessionDO) -> str:
        """Add a new session to the database."""
        with self.get_db_session() as db_session:
            session_orm = SessionORM(**session_do.model_dump())
            db_session.add(session_orm)
            db_session.commit()
            return session_orm.session_id

    def update_session(self, session_do: SessionDO) -> str:
        """Update an existing session to the database."""
        try:
            with self.get_db_session() as db_session:
                session_orm = db_session.query(SessionORM).filter(
                    SessionORM.session_id == session_do.session_id).first()
                if session_orm:
                    update_data = session_do.model_dump(exclude_unset=True)
                    for key, value in update_data.items():
                        setattr(session_orm, key, value)
                    db_session.commit()
                    return session_orm.session_id
        except Exception as e:
            LOGGER.error("session library update failed, exception: {}", e)
            db_session.rollback()

    def delete_session(self, session_id: str) -> str | None:
        """Delete a session from the database  using the provided `session_id`."""
        with self.get_db_session() as db_session:
            session_orm = db_session.query(SessionORM).filter(
                SessionORM.session_id == session_id).one_or_none()
            if session_orm:
                db_session.delete(session_orm)
                db_session.commit()
                return session_orm.session_id
            else:
                return None

    def get_session_list(self, agent_id: str) -> list[SessionDO]:
        """Get a list of sessions from the database using the provided `agent_id`."""
        with self.get_db_session() as db_session:
            session_orm_list = db_session.query(SessionORM).filter(
                SessionORM.agent_id == agent_id).order_by(desc(SessionORM.gmt_modified)).all()
            res = []
            if session_orm_list:
                for session_orm in session_orm_list:
                    res.append(self.__session_orm_to_do(session_orm))
            return res

    def get_session_detail(self, session_id: str) -> SessionDO | None:
        """Get a session from the database using the provided `session_id`"""
        with self.get_db_session() as db_session:
            session_orm = db_session.query(SessionORM).filter(
                SessionORM.session_id == session_id).first()
            if session_orm:
                return self.__session_orm_to_do(session_orm)
            else:
                return None

    @staticmethod
    def __session_orm_to_do(session_orm: SessionORM) -> SessionDO:
        """Convert a SessionORM object to a SessionDO object."""
        session_do = SessionDO(
            session_id="",
            agent_id='',
            ext_info=dict(),
        )
        for column in session_orm.__table__.columns:
            setattr(session_do, column.name,
                    getattr(session_orm, column.name))
        return session_do



# !/usr/bin/env python3
# -*- coding:utf-8 -*-

# @Time    : 2024/10/9 19:33
# @Author  : wangchongshi
# @Email   : wangchongshi.wcs@antgroup.com
# @FileName: memory_compressor_manager.py
from agentuniverse.agent.memory.memory_compressor.memory_compressor import MemoryCompressor
from agentuniverse.base.annotation.singleton import singleton
from agentuniverse.base.component.component_enum import ComponentEnum
from agentuniverse.base.component.component_manager_base import ComponentManagerBase


@singleton
class MemoryCompressorManager(ComponentManagerBase[MemoryCompressor]):
    """A singleton manager class of the MemoryCompressor."""

    def __init__(self):
        super().__init__(ComponentEnum.MEMORY_COMPRESSOR)



# !/usr/bin/env python3
# -*- coding:utf-8 -*-

# @Time    : 2024/7/24 11:37
# @Author  : fanen.lhy
# @Email   : fanen.lhy@antgroup.com
# @FileName: query_paraphraser_manager.py

from agentuniverse.base.annotation.singleton import singleton
from agentuniverse.base.component.component_enum import ComponentEnum
from agentuniverse.base.component.component_manager_base import ComponentManagerBase
from agentuniverse.agent.action.knowledge.query_paraphraser.query_paraphraser import QueryParaphraser


@singleton
class QueryParaphraserManager(ComponentManagerBase[QueryParaphraser]):
    """A singleton manager class of the QueryParaphraser."""

    def __init__(self):
        super().__init__(ComponentEnum.QUERY_PARAPHRASER)


# !/usr/bin/env python3
# -*- coding:utf-8 -*-

# @Time    : 2024/8/23 15:33
# @Author  : wangchongshi
# @Email   : wangchongshi.wcs@antgroup.com
# @FileName: workflow_agent.py
from typing import Optional

from agentuniverse.agent.agent import Agent
from agentuniverse.agent.input_object import InputObject
from agentuniverse.base.config.component_configer.configers.agent_configer import AgentConfiger
from agentuniverse.workflow.workflow import Workflow
from agentuniverse.workflow.workflow_manager import WorkflowManager
from agentuniverse.workflow.workflow_output import WorkflowOutput


class WorkflowAgent(Agent):
    """Workflow Agent class."""
    workflow_id: Optional[str] = None

    def input_keys(self) -> list[str]:
        """Return the input keys of the Agent."""
        return []

    def output_keys(self) -> list[str]:
        """Return the output keys of the Agent."""
        return []

    def parse_input(self, input_object: InputObject, agent_input: dict) -> dict:
        """Agent parameter parsing.

        Args:
            input_object (InputObject): input parameters passed by the user.
            agent_input (dict): agent input preparsed by the agent.
        Returns:
            dict: agent input parsed from `input_object` by the user.
        """
        return agent_input

    def parse_result(self, agent_result: dict) -> dict:
        """Agent result parser.

        Args:
            agent_result(dict): The raw result of the agent.
        Returns:
            dict: The parsed result of the agent
        """
        return agent_result

    def execute(self, input_object: InputObject, agent_input: dict) -> dict:
        workflow: Workflow = WorkflowManager().get_instance_obj(self.workflow_id)
        # build and run workflow
        if not workflow or workflow.graph_config is None:
            raise Exception('Workflow graph is None, please add nodes and edges to the workflow graph.')
        workflow = workflow.build()
        workflow_output: WorkflowOutput = workflow.run(input_object.to_dict())
        print(workflow_output.workflow_node_results)
        return workflow_output.workflow_end_params

    def initialize_by_component_configer(self, component_configer: AgentConfiger) -> 'WorkflowAgent':
        super().initialize_by_component_configer(component_configer)
        self.workflow_id = (self.agent_model.profile.get('workflow_id')
                            or self.agent_model.plan.get('planner', {}).get('workflow_id'))
        return self



# !/usr/bin/env python3
# -*- coding:utf-8 -*-

# @Time    : 2024/8/27 15:52
# @Author  : wangchongshi
# @Email   : wangchongshi.wcs@antgroup.com
# @FileName: plugin_service.py
import os
from typing import List

from agentuniverse_product.base.product import Product
from agentuniverse_product.base.product_manager import ProductManager
from agentuniverse_product.base.util.yaml_util import write_yaml_file
from agentuniverse_product.service.model.plugin_dto import PluginDTO
from agentuniverse_product.service.model.tool_dto import ToolDTO
from agentuniverse_product.service.tool_service.tool_service import ToolService
from agentuniverse_product.service.util.agent_util import register_product
from agentuniverse_product.service.util.common_util import get_core_path
from agentuniverse_product.service.util.plugin_util import assemble_plugin_product_config_data, \
    parse_openapi_yaml_to_tool_bundle, validate_create_plugin_parameters
from agentuniverse_product.service.util.tool_util import parse_tool_input


class PluginService:
    """Plugin Service for aU-product."""

    @staticmethod
    def get_plugin_list() -> List[PluginDTO]:
        """Get list of plugins."""
        res = []
        product_list: List[Product] = ProductManager().get_instance_obj_list()
        if len(product_list) < 1:
            return res
        for product in product_list:
            if product.type == 'PLUGIN':
                tool_id_list = product.toolset
                tool_dto_list = []
                if len(tool_id_list) > 0:
                    for tool_id in tool_id_list:
                        tool_dto: ToolDTO = ToolService().get_tool_detail(tool_id)
                        if tool_dto is None:
                            continue
                        tool_dto_list.append(tool_dto)
                plugin_dto = PluginDTO(nickname=product.nickname, avatar=product.avatar, id=product.id,
                                       toolset=tool_dto_list, description=product.description,
                                       openapi_desc=product.openapi_desc)
                res.append(plugin_dto)
        return res

    @staticmethod
    def create_plugin_with_openapi(plugin_dto: PluginDTO) -> str:
        """Create a new plugin with OpenAPI schema.

        Args:
            plugin_dto (PluginDTO): Plugin DTO.

        Returns:
            str: Plugin ID.
        """
        # single OpenAPI schema with multiple APIs is parsed into multiple tool bundles.
        validate_create_plugin_parameters(plugin_dto)
        tool_bundles = parse_openapi_yaml_to_tool_bundle(plugin_dto.openapi_desc)
        tool_id_list = []
        index = 1
        for tool in tool_bundles:
            tool_id = plugin_dto.id + '_tool_' + str(index)
            parameters = parse_tool_input(tool)
            tool_dto = ToolDTO(id=tool_id,
                               nickname=plugin_dto.nickname + '_tool_' + str(index) if plugin_dto.nickname else '',
                               avatar=plugin_dto.avatar if plugin_dto.avatar else None,
                               parameters=parameters,
                               openapi_schema=tool
                               )
            ToolService().create_tool(tool_dto)
            tool_id_list.append(tool_id)
            index += 1
        # assemble product config data
        product_config_data = assemble_plugin_product_config_data(plugin_dto, tool_id_list)
        # write product YAML file
        product_file_name = f"{plugin_dto.id}_product"
        path = get_core_path()
        product_file_path = path / 'product' / 'plugin' / f"{product_file_name}.yaml" if path\
            else os.path.join('..', '..', 'platform', 'difizen', 'product', 'plugin', f"{product_file_name}.yaml")
        write_yaml_file(str(product_file_path), product_config_data)
        register_product(str(product_file_path))
        return plugin_dto.id



import traceback

from flask import Flask, Response
from werkzeug.exceptions import HTTPException

from ..service_instance import ServiceInstance, ServiceNotFoundError
from .request_task import RequestTask
from .web_util import request_param, service_run_queue, make_standard_response
from ...base.util.logging.logging_util import LOGGER


app = Flask(__name__)
app.config['JSON_AS_ASCII'] = False
app.json.ensure_ascii = False


@app.route("/echo")
def echo():
    return 'Welcome to agentUniverse!!!'


@app.route("/liveness")
def liveness():
    return make_standard_response(success=True,
                                  result="liveness health check pass!")


@app.route("/service_run", methods=['POST'])
@request_param
def service_run(service_id: str, params: dict, saved: bool = False):
    """Synchronous invocation of an agent service.

    Request Args:
        service_id(`str`): The id of the agent service.
        params(`dict`): Json style params passed to service.
        saved(`bool`): Save the request and result into database.

    Return:
        Returns a dict containing two keys: success and result.
        success: This key holds a boolean value indicating the task was
            successfully or not.
        result: This key points to a nested dictionary that includes the
            result of the task.
    """
    params = {} if params is None else params
    request_task = RequestTask(ServiceInstance(service_id).run, saved, **params)
    result = request_task.run()
    return make_standard_response(success=True, result=result,
                                  request_id=request_task.request_id)


@app.route("/service_run_stream", methods=['POST'])
@request_param
def service_run_stream(service_id: str, params: dict, saved: bool = False):
    """Synchronous invocation of an agent service, return in stream form.

    Request Args:
        service_id(`str`): The id of the agent service.
        params(`dict`): Json style params passed to service.
        saved(`bool`): Save the request and result into database.

    Return:
        A SSE(Server-Sent Event) stream.
    """
    params = {} if params is None else params
    params['service_id'] = service_id
    task = RequestTask(service_run_queue, saved, **params)
    response = Response(task.stream_run(), mimetype="text/event-stream")
    response.headers['X-Request-ID'] = task.request_id
    return response


@app.route("/service_run_async", methods=['POST'])
@request_param
def service_run_async(service_id: str, params: dict, saved: bool = True):
    """Async invocation of an agent service, return the request id used to
    get result later.

    Request Args:
        service_id(`str`): The id of the agent service.
        params(`dict`): Json style params passed to service.
        saved(`bool`): Save the request and result into database.

    Return:
        Returns a dict containing two keys: success and result.
        success: This key holds a boolean value indicating the task was
            successfully or not.
        request_id: Stand for a single request taski, can be used in
            service_run_result api to get the result of async task.
    """
    params = {} if params is None else params
    params['service_id'] = service_id
    task = RequestTask(service_run_queue, saved, **params)
    task.async_run()
    return make_standard_response(success=True,
                                  request_id=task.request_id)


@app.route("/service_run_result", methods=['GET'])
@request_param
def service_run_result(request_id: str):
    """Get the async service result.

    Request Args:
        request_id(`str`): Request id returned by async run api.

    Return:
        Returns a dict containing two keys: success and result if request_id
        exists in database.
        success: This key holds a boolean value indicating the task was
            successfully or not.
        result: This key points to a nested dictionary that includes the
            result of the task.
    """
    data = RequestTask.query_request_state(request_id)
    if data is None:
        return make_standard_response(
            success=False,
            message=f"request {request_id} not found"
        )
    return make_standard_response(success=True, result=data,
                                  request_id=request_id)


@app.errorhandler(HTTPException)
def handle_http_exception(e):
    """A global exception handler handle flask origin http exceptions."""
    response = e.get_response()
    return response


@app.errorhandler(Exception)
def handle_exception(e):
    """A global non http exception handler"""
    LOGGER.error(traceback.format_exc())
    if isinstance(e, ServiceNotFoundError):
        return make_standard_response(success=False,
                                      message=str(e),
                                      status_code=404)
    return make_standard_response(success=False,
                                  message="Internal Server Error",
                                  status_code=500)



from typing import Optional

from .service_configer import ServiceConfiger
from ..agent.agent import Agent
from ..base.config.application_configer.application_config_manager import (
    ApplicationConfigManager
)
from ..base.component.component_base import ComponentBase
from ..base.component.component_enum import ComponentEnum


class Service(ComponentBase):
    """The basic class of the service."""

    # Basic attributes of the service class.
    component_type: ComponentEnum = ComponentEnum.SERVICE
    name: Optional[str] = None
    description: Optional[str] = None
    agent: Optional[Agent] = None

    def __post_init_post_parse__(self):
        """Init service code with service name."""
        self.__service_code: Optional[str] = self.get_instance_code()

    def get_instance_code(self) -> str:
        """Generate the full service code from service name. """
        app_cfg_manager: ApplicationConfigManager = ApplicationConfigManager()
        appname = app_cfg_manager.app_configer.base_info_appname
        return f"{appname}.service.{self.name}"

    def initialize_by_component_configer(self,
                                         service_configer: ServiceConfiger) \
            -> 'Service':
        """Initialize the Service by the ComponentConfiger object.

        Args:
            service_configer(ServiceConfiger): A configer contains service
            basic info.
        Returns:
            Service: A Service instance.
        """
        self.name = service_configer.name
        self.description = service_configer.description
        self.agent = service_configer.agent
        return self

    def run(self, **kwargs) -> str:
        """The executed function when the service is called."""
        return self.agent.run(**kwargs).to_json_str()

    @property
    def service_code(self):
        """The unique code of each service, generate from service name."""
        return self.__service_code



# !/usr/bin/env python3
# -*- coding:utf-8 -*-

# @Time    : 2024/6/3 23:09
# @Author  : fanen.lhy
# @Email   : fanen.lhy@antgroup.com
# @FileName: post_fork_queue.py

from typing import List, Tuple, Callable, Any

FunctionWithArgs = Tuple[Callable, Tuple[Any, ...], dict]
POST_FORK_QUEUE: List[FunctionWithArgs] = []


def add_post_fork(func: Callable, *args: Any, **kwargs: Any) -> None:
    """
    Add funcs and parameters into a waiting list, all of them will be executed
    after gunicorn worker child processes have been forked, or before flask
    main app start if you work without gunicorn.
    """

    POST_FORK_QUEUE.append((func, args, kwargs))






from pathlib import Path
from typing import List, Optional, Dict

from agentuniverse.agent.action.knowledge.reader.reader import Reader
from agentuniverse.agent.action.knowledge.store.document import Document


class LineTxtReader(Reader):

    def _load_data(self, fpath: Path, ext_info: Optional[Dict] = None) -> List[Document]:
        dlist = []

        with open(fpath, 'r', encoding='utf-8') as file:

            metadata = {"file_name": file.name}
            if ext_info is not None:
                metadata.update(ext_info)

            for line in file:
                dlist.append(Document(text=line, metadata=metadata or {}))

        return dlist


class TxtReader(Reader):
    """Txt reader."""

    def _load_data(self, fpath: Path, ext_info: Optional[Dict] = None) -> List[Document]:

        with open(fpath, 'r', encoding='utf-8') as file:

            metadata = {"file_name": file.name}
            if ext_info is not None:
                metadata.update(ext_info)

            txt = file.read()

        return [Document(text=txt, metadata=metadata or {})]



# !/usr/bin/env python3
# -*- coding:utf-8 -*-

# @Time    : 2024/8/20 20:01
# @Author  : wangchongshi
# @Email   : wangchongshi.wcs@antgroup.com
# @FileName: tool_node.py
from typing import List, Optional

from agentuniverse.agent.action.tool.tool import Tool
from agentuniverse.agent.action.tool.tool_manager import ToolManager
from agentuniverse.workflow.node.enum import NodeEnum, NodeStatusEnum
from agentuniverse.workflow.node.node import NodeData, Node
from agentuniverse.workflow.node.node_output import NodeOutput
from agentuniverse.workflow.node.node_config import ToolNodeInputParams, NodeInfoParams, NodeOutputParams
from agentuniverse.workflow.workflow_output import WorkflowOutput


class ToolNodeData(NodeData):
    inputs: Optional[ToolNodeInputParams] = None


class ToolNode(Node):
    """The basic class of the tool node."""
    _data_cls = ToolNodeData

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.type = NodeEnum.TOOL

    def _run(self, workflow_output: WorkflowOutput) -> NodeOutput:
        inputs: ToolNodeInputParams = self._data.inputs
        tool_params: List[NodeInfoParams] = inputs.tool_param
        tool_id = None
        for tool_param in tool_params:
            if tool_param.name == 'id':
                if isinstance(tool_param.value, dict):
                    tool_id = tool_param.value['content']
                else:
                    tool_id = tool_param.value
        tool: Tool = ToolManager().get_instance_obj(tool_id)
        if tool is None:
            raise ValueError("No tool with id {} was found.".format(tool_id))

        tool_input_params = self._resolve_input_params(inputs.input_param, workflow_output)
        tool_output = tool.run(**tool_input_params)
        output_params: List[NodeOutputParams] = self._data.outputs

        if isinstance(tool_output, str):
            output_params[0].value = tool_output
        elif isinstance(tool_output, dict):
            for output_param in output_params:
                output_param.value = tool_output.get(output_param.name, None)
        else:
            raise TypeError(f"The type of tool_output is not supported.")
        workflow_output.workflow_parameters[self.id] = output_params
        return NodeOutput(node_id=self.id, status=NodeStatusEnum.SUCCEEDED, result=output_params)



# !/usr/bin/env python3
# -*- coding:utf-8 -*-

# @Time    : 2024/3/14 11:54
# @Author  : jerry.zzw 
# @Email   : jerry.zzw@antgroup.com
# @FileName: config_manager.py
from typing import Optional
from agentuniverse.base.annotation.singleton import singleton
from agentuniverse.base.config.application_configer.app_configer import AppConfiger


@singleton
class ApplicationConfigManager(object):
    """The ConfigManager class, which is used to load and manage the configuration."""

    def __init__(self):
        """Initialize the ConfigManager."""
        self.__app_configer: Optional[AppConfiger] = None

    @property
    def app_configer(self):
        """Return the AppConfiger object."""
        if self.__app_configer is None:
            raise ValueError("The AppConfiger object is not set.")
        return self.__app_configer

    @app_configer.setter
    def app_configer(self, app_configer: AppConfiger):
        """Set the AppConfiger object."""
        self.__app_configer = app_configer




# !/usr/bin/env python3
# -*- coding:utf-8 -*-

# @Time    : 2024/3/13 13:54
# @Author  : wangchongshi
# @Email   : wangchongshi.wcs@antgroup.com
# @FileName: tool_manager.py
from agentuniverse.base.annotation.singleton import singleton
from agentuniverse.base.component.component_enum import ComponentEnum
from agentuniverse.base.component.component_manager_base import ComponentManagerBase


@singleton
class ToolManager(ComponentManagerBase):
    """The ToolManager class, which is used to manage the tools."""

    def __init__(self):
        """Initialize the ToolManager."""
        super().__init__(ComponentEnum.TOOL)



# !/usr/bin/env python3
# -*- coding:utf-8 -*-

# @Time    : 2024/8/28 14:29
# @Author  : wangchongshi
# @Email   : wangchongshi.wcs@antgroup.com
# @FileName: api_tool.py
import json
from typing import Any, Optional
from urllib.parse import urlencode
import httpx

from agentuniverse.agent.action.tool.tool import Tool, ToolInput
from agentuniverse.agent.action.tool.utils import ssrf_proxy
from agentuniverse.base.config.component_configer.configers.tool_configer import ToolConfiger


class APITool(Tool):
    """The basic class for api tool model.

    Attributes:
        openapi_spec(str): The openapi schema of the api tool.
    """
    openapi_spec: Optional[dict] = None

    def execute(self, tool_input: ToolInput):
        res = self.do_http_request(self.openapi_spec.get('url'), self.openapi_spec.get('method'), {},
                                   tool_input.to_dict())
        return res.text

    def initialize_by_component_configer(self, component_configer: ToolConfiger) -> 'APITool':
        super().initialize_by_component_configer(component_configer)
        if component_configer.__dict__['openapi_spec']:
            self.openapi_spec = component_configer.__dict__['openapi_spec']
        return self

    def convert_body_property_any_of(self, property: dict[str, Any], value: Any, any_of: list[dict[str, Any]],
                                     max_recursive=10) -> Any:
        """Convert a property value based on its anyOf type."""
        if max_recursive <= 0:
            raise Exception("Max recursion depth reached")
        for option in any_of or []:
            try:
                if 'type' in option:
                    # Attempt to convert the value based on the type.
                    if option['type'] == 'integer' or option['type'] == 'int':
                        return int(value)
                    elif option['type'] == 'number':
                        if '.' in str(value):
                            return float(value)
                        else:
                            return int(value)
                    elif option['type'] == 'string':
                        return str(value)
                    elif option['type'] == 'boolean':
                        if str(value).lower() in ['true', '1']:
                            return True
                        elif str(value).lower() in ['false', '0']:
                            return False
                        else:
                            # Not a boolean, try next option
                            continue
                    elif option['type'] == 'null' and not value:
                        return None
                    else:
                        # Unsupported type, try next option
                        continue
                elif 'anyOf' in option and isinstance(option['anyOf'], list):
                    # Recursive call to handle nested anyOf
                    return self.convert_body_property_any_of(property, value, option['anyOf'], max_recursive - 1)
            except ValueError:
                # Conversion failed, try next option
                continue
        return value

    def convert_body_property_type(self, property: dict[str, Any], value: Any) -> Any:
        """Convert a property value based on its type."""
        try:
            if 'type' in property:
                if property['type'] == 'integer' or property['type'] == 'int':
                    return int(value)
                elif property['type'] == 'number':
                    # check if it is a float
                    if '.' in value:
                        return float(value)
                    else:
                        return int(value)
                elif property['type'] == 'string':
                    return str(value)
                elif property['type'] == 'boolean':
                    return bool(value)
                elif property['type'] == 'null':
                    if value is None:
                        return None
                elif property['type'] == 'object':
                    if isinstance(value, str):
                        try:
                            return json.loads(value)
                        except ValueError:
                            return value
                    elif isinstance(value, dict):
                        return value
                    else:
                        return value
                else:
                    raise ValueError(
                        f"Invalid type {property['type']} for property {property}")
            elif 'anyOf' in property and isinstance(property['anyOf'], list):
                return self.convert_body_property_any_of(property, value, property['anyOf'])
        except ValueError as e:
            return value

    def do_http_request(self, url: str, method: str, headers: dict[str, Any],
                        parameters: dict[str, Any]) -> httpx.Response:
        """Do an HTTP request.

        Args:
            url (str): The URL to request.
            method (str): The HTTP method to use.
            headers (dict[str, Any]): The headers to include in the request.
            parameters (dict[str, Any]): The parameters to include in the request.

        Returns:
            httpx.Response: The response from the request.
        """
        method = method.lower()
        params = {}
        path_params = {}
        body = {}
        cookies = {}
        # check parameters
        for parameter in self.openapi_spec.get('operation').get('parameters', []):
            value = self.get_parameter_value(parameter, parameters)
            if value is not None:
                if parameter['in'] == 'path':
                    path_params[parameter['name']] = value

                elif parameter['in'] == 'query':
                    params[parameter['name']] = value

                elif parameter['in'] == 'cookie':
                    cookies[parameter['name']] = value

                elif parameter['in'] == 'header':
                    headers[parameter['name']] = value

        # check if there is a request body and handle it
        if 'requestBody' in self.openapi_spec and self.openapi_spec['requestBody'] is not None:
            # handle json request body
            if 'content' in self.openapi_spec['requestBody']:
                for content_type in self.openapi_spec['requestBody']['content']:
                    headers['Content-Type'] = content_type
                    body_schema = self.openapi_spec[
                        'requestBody']['content'][content_type]['schema']
                    required = body_schema.get('required', [])
                    properties = body_schema.get('properties', {})
                    for name, property in properties.items():
                        if name in parameters:
                            # convert type
                            body[name] = self.convert_body_property_type(
                                property, parameters[name])
                        elif name in required:
                            raise Exception(
                                f"Missing required parameter {name} in operation {self.plugin_api_model.operation_id}"
                            )
                        elif 'default' in property:
                            body[name] = property['default']
                        else:
                            body[name] = None
                    break

        # replace path parameters
        for name, value in path_params.items():
            url = url.replace(f'{{{name}}}', f'{value}')

        # parse http body data if needed, for GET/HEAD/OPTIONS/TRACE
        if 'Content-Type' in headers:
            if headers['Content-Type'] == 'application/json':
                body = json.dumps(body)
            elif headers['Content-Type'] == 'application/x-www-form-urlencoded':
                body = urlencode(body)
            else:
                body = body
        if method in ('get', 'head', 'post', 'put', 'delete', 'patch'):
            response = getattr(ssrf_proxy, method)(url, params=params, headers=headers, data=body,
                                                   follow_redirects=True)
            return response
        else:
            raise ValueError(f'Invalid http method')

    @staticmethod
    def get_parameter_value(parameter, parameters):
        if parameter['name'] in parameters:
            return parameters[parameter['name']]
        elif parameter.get('required', False):
            raise Exception(f"Missing required parameter {parameter['name']}")
        else:
            return (parameter.get('schema', {}) or {}).get('default', None)

    @staticmethod
    def validate_and_parse_response(response: httpx.Response) -> str:
        """Validate and parse the response from the tool response."""
        if isinstance(response, httpx.Response):
            if response.status_code >= 400:
                raise Exception(
                    f"Request failed with status code {response.status_code} and {response.text}")
            if not response.content:
                return 'Empty response from the tool, please check your parameters and try again.'
            try:
                response = response.json()
                try:
                    return json.dumps(response, ensure_ascii=False)
                except Exception as e:
                    return json.dumps(response)
            except Exception as e:
                return response.text
        else:
            raise ValueError(f'Invalid response type {type(response)}')



# !/usr/bin/env python3
# -*- coding:utf-8 -*-

# @Time    : 2024/3/13 14:29
# @Author  : wangchongshi
# @Email   : wangchongshi.wcs@antgroup.com
# @FileName: tool.py
from abc import abstractmethod
import json
from typing import List, Optional

from pydantic import BaseModel
from langchain.tools import Tool as LangchainTool

from agentuniverse.agent.action.tool.enum import ToolTypeEnum
from agentuniverse.base.annotation.trace import trace_tool
from agentuniverse.base.component.component_base import ComponentBase
from agentuniverse.base.component.component_enum import ComponentEnum
from agentuniverse.base.config.application_configer.application_config_manager import ApplicationConfigManager
from agentuniverse.base.config.component_configer.configers.tool_configer import ToolConfiger


class ToolInput(BaseModel):
    """The basic class for tool input."""

    def __init__(self, params: dict, **kwargs):
        super().__init__(**kwargs)
        self.__origin_params = params
        for k, v in params.items():
            self.__dict__[k] = v

    def to_dict(self):
        return self.__origin_params

    def to_json_str(self):
        return json.dumps(self.__origin_params, ensure_ascii=False)

    def add_data(self, key, value):
        self.__origin_params[key] = value
        self.__dict__[key] = value

    def get_data(self, key, default=None):
        return self.__origin_params.get(key, default)


class Tool(ComponentBase):
    """
    The basic class for tool model.

    Attributes:
        name (str): The name of the tool.
        description (str): The description of the tool.
        tool_type (ToolTypeEnum): The type of the tool.
        input_keys (Optional[List]): The input keys of the tool, e.g. ['input1', 'input2']
    """

    name: str = ""
    description: Optional[str] = None
    tool_type: ToolTypeEnum = ToolTypeEnum.FUNC
    input_keys: Optional[List] = None

    def __init__(self, **kwargs):
        super().__init__(component_type=ComponentEnum.TOOL, **kwargs)

    @trace_tool
    def run(self, **kwargs):
        """The callable method that runs the tool."""
        self.input_check(kwargs)
        tool_input = ToolInput(kwargs)
        return self.execute(tool_input)

    def input_check(self, kwargs: dict) -> None:
        """Check whether the input parameters of the tool contain input keys of the tool"""
        if self.input_keys:
            for key in self.input_keys:
                if key not in kwargs.keys():
                    raise Exception(f'{self.get_instance_code()} - The input must include key: {key}.')

    @trace_tool
    def langchain_run(self, *args, callbacks=None, **kwargs):
        """The callable method that runs the tool."""
        kwargs["callbacks"] = callbacks
        tool_input = ToolInput(kwargs)
        parse_result = self.parse_react_input(args[0])
        for key in self.input_keys:
            tool_input.add_data(key, parse_result[key])
        return self.execute(tool_input)

    def parse_react_input(self, input_str: str):
        """
            parse react string to you input
            you can define your own logic here by override this function
        """
        return {
            self.input_keys[0]: input_str
        }

    @abstractmethod
    def execute(self, tool_input: ToolInput):
        raise NotImplementedError

    def as_langchain(self) -> LangchainTool:
        """Convert the agentUniverse(aU) tool class to the langchain tool class."""
        return LangchainTool(name=self.name,
                             func=self.langchain_run,
                             description=self.description)

    def get_instance_code(self) -> str:
        """Return the full name of the tool."""
        appname = ApplicationConfigManager().app_configer.base_info_appname
        return f'{appname}.{self.component_type.value.lower()}.{self.name}'

    def initialize_by_component_configer(self, component_configer: ToolConfiger) -> 'Tool':
        """Initialize the LLM by the ComponentConfiger object.
        Args:
            component_configer(LLMConfiger): the ComponentConfiger object
        Returns:
            LLM: the LLM object
        """
        if component_configer.name:
            self.name = component_configer.name
        if component_configer.description:
            self.description = component_configer.description
        if component_configer.tool_type:
            self.tool_type = next((member for member in ToolTypeEnum if member.value == component_configer.tool_type))
        if component_configer.input_keys:
            self.input_keys = component_configer.input_keys
        return self



# !/usr/bin/env python3
# -*- coding:utf-8 -*-

# @Time    : 2024/7/26 14:26
# @Author  : wangchongshi
# @Email   : wangchongshi.wcs@antgroup.com
# @FileName: session_dto.py
from typing import Optional

from pydantic import BaseModel, Field

from agentuniverse_product.service.model.message_dto import MessageDTO


class SessionDTO(BaseModel):
    id: str = Field(description="ID")
    agent_id: str = Field(description="session agent id")
    messages: Optional[list[MessageDTO]] = Field(description="session messages", default=[])
    gmt_created: Optional[str] = Field(description="session create time")
    gmt_modified: Optional[str] = Field(description="session update time")



# !/usr/bin/env python3
# -*- coding:utf-8 -*-
# @Time    : 2024/3/12 15:13
# @Author  : heji
# @Email   : lc299034@antgroup.com
# @FileName: agent_model.py
"""Agent model class."""
from typing import Optional
from pydantic import BaseModel


class AgentModel(BaseModel):
    """The parent class of all agent models, containing only attributes."""

    info: Optional[dict] = dict()
    profile: Optional[dict] = dict()
    plan: Optional[dict] = dict()
    memory: Optional[dict] = dict()
    action: Optional[dict] = dict()
    work_pattern: Optional[dict] = dict()

    def llm_params(self) -> dict:
        """
        Returns:
            dict: The parameters for the LLM.
        """
        params = {}
        for key, value in self.profile.get('llm_model').items():
            if key == 'name' or key == 'prompt_processor':
                continue
            if key == 'model_name':
                params['model'] = value
            else:
                params[key] = value
        return params



# !/usr/bin/env python3
# -*- coding:utf-8 -*-

# @Time    : 2024/8/23 16:37
# @Author  : wangchongshi
# @Email   : wangchongshi.wcs@antgroup.com
# @FileName: node_constant.py
from agentuniverse.workflow.node.agent_node import AgentNode
from agentuniverse.workflow.node.condition_node import ConditionNode
from agentuniverse.workflow.node.end_node import EndNode
from agentuniverse.workflow.node.knowledge_node import KnowledgeNode
from agentuniverse.workflow.node.llm_node import LLMNode
from agentuniverse.workflow.node.start_node import StartNode
from agentuniverse.workflow.node.tool_node import ToolNode

NODE_CLS_MAPPING = {
    "start": StartNode,
    "end": EndNode,
    "tool": ToolNode,
    "knowledge": KnowledgeNode,
    "agent": AgentNode,
    "llm": LLMNode,
    "ifelse": ConditionNode,
}




# !/usr/bin/env python3
# -*- coding:utf-8 -*-

# @Time    : 2024/8/21 14:24
# @Author  : wangchongshi
# @Email   : wangchongshi.wcs@antgroup.com
# @FileName: workflow_output.py
from typing import Optional, Dict, Any, List

from pydantic import BaseModel

from agentuniverse.workflow.node.node_config import NodeOutputParams
from agentuniverse.workflow.node.node_output import NodeOutput


class WorkflowOutput(BaseModel):
    """The basic class of the workflow output."""

    workflow_id: str = None
    metadata: Optional[Dict[str, Any]] = dict()
    workflow_parameters: Optional[Dict[str, List[NodeOutputParams]]] = dict()
    workflow_node_results: Optional[Dict[str, NodeOutput]] = dict()
    workflow_start_params: Optional[Dict[str, Any]] = dict()
    workflow_end_params: Optional[Dict[str, Any]] = dict()



# !/usr/bin/env python3
# -*- coding:utf-8 -*-

# @Time    : 2024/3/18 14:21
# @Author  : wangchongshi
# @Email   : wangchongshi.wcs@antgroup.com
# @FileName: pdf_parser.py
from pathlib import Path
from typing import List, Optional, Dict, Union

from agentuniverse.agent.action.knowledge.reader.reader import Reader
from agentuniverse.agent.action.knowledge.store.document import Document


class PdfReader(Reader):
    """PDF reader."""

    def _load_data(self, file: Union[str, Path], ext_info: Optional[Dict] = None) -> List[Document]:
        """Parse the pdf file.

        Note:
            `pypdf` is required to read PDF files: `pip install pypdf`
        """
        try:
            import pypdf
        except ImportError:
            raise ImportError(
                "pypdf is required to read PDF files: `pip install pypdf`"
            )
        if isinstance(file, str):
            file = Path(file)
        with open(file, "rb") as fp:
            # Create a PDF object
            pdf = pypdf.PdfReader(fp)

            # Get the number of pages in the PDF document
            num_pages = len(pdf.pages)

            # Iterate over every page
            docs = []
            for page in range(num_pages):
                # Extract the text from the page
                page_text = pdf.pages[page].extract_text()
                page_label = pdf.page_labels[page]

                metadata = {"page_label": page_label, "file_name": file.name}
                if ext_info is not None:
                    metadata.update(ext_info)

                docs.append(Document(text=page_text, metadata=metadata))
            return docs



# !/usr/bin/env python3
# -*- coding:utf-8 -*-

# @Time    : 2024/5/21 13:52
# @Author  : weizjajj
# @Email   : weizhongjie.wzj@antgroup.com
# @FileName: openai_style_llm.py

from typing import Any, Optional, AsyncIterator, Iterator, Union

import httpx
import openai
import tiktoken
from langchain_core.language_models.base import BaseLanguageModel
from openai import OpenAI, AsyncOpenAI

from agentuniverse.base.config.component_configer.configers.llm_configer import LLMConfiger
from agentuniverse.base.util.env_util import get_from_env
from agentuniverse.llm.llm import LLM, LLMOutput
from agentuniverse.llm.openai_style_langchain_instance import LangchainOpenAIStyleInstance


class OpenAIStyleLLM(LLM):
    """This is a wrapper around the OpenAI API that implements a chat interface for the LLM.

    It uses the `chat` endpoint of the OpenAI API.
    It also supports using the `completion` endpoint instead of the `chat` endpoint.
    It supports both sync and async modes.

    Attributes:
        api_key (Optional[str]): The API key to use for authentication.
        organization (Optional[str]): The organization ID to use for authentication.
        api_base (Optional[str]): The base URL to use for the API requests.
        proxy (Optional[str]): The proxy to use for the API requests.
        client_args (Optional[dict]): Additional arguments to pass to the client.
    """

    api_key: Optional[str] = None
    organization: Optional[str] = None
    api_base: Optional[str] = None
    proxy: Optional[str] = None
    client_args: Optional[dict] = None

    def _new_client(self):
        """Initialize the openai client."""
        if self.client is not None:
            return self.client
        return OpenAI(
            api_key=self.api_key,
            organization=self.organization,
            base_url=self.api_base,
            timeout=self.request_timeout,
            max_retries=self.max_retries,
            http_client=httpx.Client(proxy=self.proxy) if self.proxy else None,
            **(self.client_args or {}),
        )

    def _new_async_client(self):
        """Initialize the openai async client."""
        if self.async_client is not None:
            return self.async_client
        return AsyncOpenAI(
            api_key=self.api_key,
            organization=self.organization,
            base_url=self.api_base,
            timeout=self.request_timeout,
            max_retries=self.max_retries,
            http_client=httpx.AsyncClient(proxy=self.proxy) if self.proxy else None,
            **(self.client_args or {}),
        )

    def _call(self, messages: list, **kwargs: Any) -> Union[LLMOutput, Iterator[LLMOutput]]:
        """Run the OpenAI LLM.

        Args:
            messages (list): The messages to send to the LLM.
            **kwargs: Arbitrary keyword arguments.
        """
        streaming = kwargs.pop("streaming") if "streaming" in kwargs else self.streaming
        if 'stream' in kwargs:
            streaming = kwargs.pop('stream')
        self.client = self._new_client()
        client = self.client
        chat_completion = client.chat.completions.create(
            messages=messages,
            model=kwargs.pop('model', self.model_name),
            temperature=kwargs.pop('temperature', self.temperature),
            stream=kwargs.pop('stream', streaming),
            max_tokens=kwargs.pop('max_tokens', self.max_tokens),
            **kwargs,
        )
        if not streaming:
            text = chat_completion.choices[0].message.content
            return LLMOutput(text=text, raw=chat_completion.model_dump())
        return self.generate_stream_result(chat_completion)

    async def _acall(self, messages: list, **kwargs: Any) -> Union[LLMOutput, AsyncIterator[LLMOutput]]:
        """Asynchronously run the OpenAI LLM.

        Args:
            messages (list): The messages to send to the LLM.
            **kwargs: Arbitrary keyword arguments.
        """
        streaming = kwargs.pop("streaming") if "streaming" in kwargs else self.streaming
        if 'stream' in kwargs:
            streaming = kwargs.pop('stream')
        self.async_client = self._new_async_client()
        async_client = self.async_client
        chat_completion = await async_client.chat.completions.create(
            messages=messages,
            model=kwargs.pop('model', self.model_name),
            temperature=kwargs.pop('temperature', self.temperature),
            stream=kwargs.pop('stream', streaming),
            max_tokens=kwargs.pop('max_tokens', self.max_tokens),
            **kwargs,
        )
        if not streaming:
            text = chat_completion.choices[0].message.content
            return LLMOutput(text=text, raw=chat_completion.model_dump())
        return self.agenerate_stream_result(chat_completion)

    def as_langchain(self) -> BaseLanguageModel:
        """Convert the AgentUniverse(AU) openai llm class to the langchain openai llm class."""
        return LangchainOpenAIStyleInstance(self)

    def set_by_agent_model(self, **kwargs) -> None:
        """ Assign values of parameters to the OpenAILLM model in the agent configuration."""
        copied_obj = super().set_by_agent_model(**kwargs)
        if 'api_key' in kwargs and kwargs['api_key']:
            copied_obj.api_key = kwargs['api_key']
        if 'api_base' in kwargs and kwargs['api_base']:
            copied_obj.api_base = kwargs['api_base']
        if 'proxy' in kwargs and kwargs['proxy']:
            copied_obj.proxy = kwargs['proxy']
        if 'client_args' in kwargs and kwargs['client_args']:
            copied_obj.client_args = kwargs['client_args']
        return copied_obj


    @staticmethod
    def parse_result(chunk):
        """Generate the result of the stream."""
        chat_completion = chunk
        if not isinstance(chunk, dict):
            chunk = chunk.dict()
        if len(chunk["choices"]) == 0:
            return
        choice = chunk["choices"][0]
        message = choice.get("delta")
        text = message.get("content")
        if not text:
            return
        return LLMOutput(text=text, raw=chat_completion.model_dump())

    def generate_stream_result(self, stream: openai.Stream):
        """Generate the result of the stream."""
        for chunk in stream:
            llm_output = self.parse_result(chunk)
            if llm_output:
                yield llm_output

    async def agenerate_stream_result(self, stream: AsyncIterator) -> AsyncIterator[LLMOutput]:
        """Generate the result of the stream."""
        async for chunk in stream:
            llm_output = self.parse_result(chunk)
            if llm_output:
                yield llm_output

    def initialize_by_component_configer(self, component_configer: LLMConfiger) -> 'LLM':
        if 'api_base' in component_configer.configer.value:
            self.api_base = component_configer.configer.value.get('api_base')
        elif 'api_base_env' in component_configer.configer.value:
            self.api_base = get_from_env(component_configer.configer.value.get('api_base_env'))
        if 'api_key' in component_configer .configer.value:
            self.api_key = component_configer.configer.value.get('api_key')
        elif 'api_key_env' in component_configer.configer.value:
            self.api_key = get_from_env(component_configer.configer.value.get('api_key_env'))
        return super().initialize_by_component_configer(component_configer)

    def get_num_tokens(self, text: str) -> int:
        """Get the number of tokens present in the text.

        Useful for checking if an input will fit in an openai model's context window.

        Args:
            text: The string input to tokenize.

        Returns:
            The integer number of tokens in the text.
        """
        try:
            encoding = tiktoken.encoding_for_model(self.model_name)
        except KeyError:
            encoding = tiktoken.get_encoding("cl100k_base")
        return len(encoding.encode(text))

    def max_context_length(self) -> int:
        """Return the maximum length of the context."""
        if super().max_context_length():
            return super().max_context_length()
